#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import logging
import os
import multiprocessing
import time
import pandas as pd
import numpy as np

from wiliot_core import WiliotGateway, ActionType, DataType, set_logger
from wiliot_testers.sample.modules.resolve_packets import ResolvePackets, TagStatus
from wiliot_testers.utils.wiliot_external_ids import is_external_id_valid
from wiliot_testers.association_tester.configs.packets_configs import get_packet_param
pd.options.mode.chained_assignment = None  # default='warn'

RESET_GW_TIME = 300  # number of seconds we allow to not getting packets until we reset the GW
RESOLVE_Q_SIZE = 100


class ResolverProcess(ResolvePackets):
    def __init__(self, owner_id, env, need_to_resolve_q, resolved_q, stop_event):
        logger_path, self.logger = set_logger('ResolverProcess', 'resolver_process', 'resolver')
        self.resolved_q = resolved_q
        super().__init__(tags_in_test=[], owner_id=owner_id, env=env, resolve_q=need_to_resolve_q,
                         set_tags_status_df=self.add_to_resolved_q, stop_event_trig=stop_event,
                         logger_name=self.logger.name)
        self.run()

    def check_tag_status(self, ex_id):
        if is_external_id_valid(ex_id['externalId']):
            status = TagStatus.INSIDE_TEST
        else:
            status = TagStatus.OUT_INVALID
            self.logger.warning(f'found invalid external id: {ex_id["externalId"]}, check serialization')
        self.logger.info(f'found new tag in test: {ex_id}')
        return status

    def add_to_resolved_q(self, element):
        self.resolved_q.put(element, block=False)


class ReelVerification(object):
    def __init__(self, user_input, stop_event, is_app_running, logger_name, logger_path, exception_q):
        self.logger = logging.getLogger(logger_name)
        self.logger_path = logger_path
        self.user_inputs = user_input
        self.stop_event = stop_event
        self.exception_q = exception_q
        self.is_running = False
        self.is_main_app_running = is_app_running

        # connect to the gw:
        self.gw_obj = None
        self.gw_init()
        self.gw_reset_and_config()

        # init data:
        self.packets_df = pd.DataFrame()
        # start app
        try:
            need_to_resolve_q = multiprocessing.Queue(maxsize=RESOLVE_Q_SIZE)
            resolved_q = multiprocessing.Queue(maxsize=RESOLVE_Q_SIZE)
            self.resolver_handler = multiprocessing.Process(target=ResolverProcess,
                                                            args=(self.user_inputs['owner_id'],
                                                                  self.user_inputs['env'],
                                                                  need_to_resolve_q,
                                                                  resolved_q,
                                                                  stop_event))

            self.need_to_resolve_q = need_to_resolve_q
            self.resolved_q = resolved_q

        except Exception as e:
            self.logger.warning(f'exception during ReelVerification init: {e}')
            raise e

    def gw_init(self):
        """
        initialize gw and the data listener thread
        """
        self.gw_obj = WiliotGateway(auto_connect=True, logger_name=self.logger.name,
                                    log_dir_for_multi_processes=os.path.dirname(self.logger_path))
        try:
            if self.gw_obj.connected:
                self.gw_obj.start_continuous_listener()
            else:
                raise Exception('gateway was not detected, please check connection')
        except Exception as e:
            self.logger.warning(e)
            raise e

    def gw_reset_and_config(self):
        """
        config gw to tester mode
        """
        self.gw_obj.reset_gw()
        if not self.gw_obj.is_gw_alive():
            raise Exception('GW is not responding')

        energy_pattern = int(self.user_inputs['energy_pattern'])
        time_profile = [int(x) for x in self.user_inputs['time_profile'].split(',')]
        ble_power = int(self.user_inputs['ble_power'])
        sub1g_power = int(self.user_inputs['sub1g_power'])
        sub1g_freq = int(self.user_inputs['sub1g_freq'])
        scan_ch = int(self.user_inputs['scan_ch'])
        is_bridge = self.user_inputs['is_listen_bridge'].lower() == 'yes'
        self.gw_obj.start_continuous_listener()
        if is_bridge:
            self.gw_obj.write('!listen_to_tag_only 0', with_ack=True)
        self.gw_obj.write(f'!set_sub_1_ghz_energizing_frequency {sub1g_freq}', with_ack=True)
        self.gw_obj.config_gw(received_channel=scan_ch,
                              energy_pattern_val=energy_pattern,
                              time_profile_val=time_profile,
                              effective_output_power_val=ble_power,
                              sub1g_output_power_val=sub1g_power,
                              start_gw_app=True, with_ack=True)

    def run_app(self):
        try:
            self.resolver_handler.start()
        except Exception as e:
            self.logger.warning(f'exception during ReelVerification run_app: {e}')
            raise e
        self.run()

    def run(self):
        self.logger.info('GetPackets Start')
        self.is_running = True
        cur_time = time.time()
        while True:
            time.sleep(0.050)
            try:
                if self.stop_event.is_set():
                    self.logger.info('GetPackets Stop')
                    self.exit_app()
                    return
                elif self.is_running != self.is_main_app_running():
                    self.is_running = not self.is_running
                    if self.is_running:
                        self.continue_app()
                    else:
                        self.pause_app()
                elif not self.exception_q.empty():
                    time.sleep(0.1)
                    continue

                if self.is_running:
                    # check if there is data to read
                    if self.gw_obj.is_data_available():
                        cur_time = time.time()
                        tag_collections_in = self.gw_obj.get_packets(action_type=ActionType.ALL_SAMPLE,
                                                                     data_type=DataType.TAG_COLLECTION)
                        for tag in tag_collections_in.keys():
                            if tag in self.packets_df.index:
                                self.packets_df.loc[tag, 'n_packets'] += 1
                            else:
                                self.logger.info(f'new tag was detected: {tag}')
                                try:
                                    new_packet = tag_collections_in[tag][0]
                                    self.need_to_resolve_q.put({'tag': tag,
                                                                'payload': new_packet.get_payload()},
                                                               block=False)
                                    self.packets_df = pd.concat([self.packets_df,
                                                                 pd.DataFrame({'n_packets': 1,
                                                                               'external_id': 'unknown'},
                                                                              index=[tag])])
                                    self.update_selected_packet(new_packet, tag)
                                except Exception as e:
                                    self.logger.warning(f'could not add new data to need_to_resolve_q or to packets_df '
                                                        f'due to: {e}')
                            self.get_one_packet(tag, tag_collections_in[tag])

                    elif self.gw_obj.get_read_error_status():
                        raise Exception(f'GW got serial reading error during run. Please disconnect and reconnect the GW')

                    if not self.resolved_q.empty():
                        self.merge_resolved_data()

                    if time.time() - cur_time > RESET_GW_TIME:
                        self.logger.warning(f'did NOT get any packets after {RESET_GW_TIME} seconds. reset and config gw')
                        self.gw_reset_and_config()
                        cur_time = time.time()
                else:
                    time.sleep(1)
                    cur_time = time.time()

            except Exception as e:
                self.logger.warning(f'GetPackets got exception: {e}')
                if not self.exception_q.full():
                    self.exception_q.put(f'GetPackets: {e}')

    def update_selected_packet(self, new_packet, tag, ind=0):
        self.packets_df.loc[tag, 'encrypted_packet'] = new_packet.get_packet_content(get_gw_data=True,
                                                                                     sprinkler_num=ind)
        self.packets_df.loc[tag, 'rssi'] = np.take(new_packet.gw_data['rssi'], ind)
        self.packets_df.loc[tag, 'packet_type'] = new_packet.packet_data['decrypted_packet_type']

    def get_one_packet(self, tag, packet_list):

        for p in packet_list:
            desired_packet_type = get_packet_param(p.packet_data['flow_ver'], 'packet_type')
            if p.packet_data['decrypted_packet_type'] != desired_packet_type:
                continue
            if self.packets_df.loc[tag, 'packet_type'] != desired_packet_type:
                self.update_selected_packet(p, tag)

            for ind in range(len(p)):
                rssi = np.take(p.gw_data['rssi'], ind)
                if rssi < self.packets_df.loc[tag, 'rssi']:
                    self.update_selected_packet(p, tag, ind)

    def merge_resolved_data(self):
        n = self.resolved_q.qsize()
        for _ in range(n):
            resolved = self.resolved_q.get(block=False)
            self.packets_df.loc[resolved['adv_address'], 'external_id'] = resolved['external_id']

    def pause_app(self):
        self.gw_obj.reset_gw()
        time.sleep(1)
        self.gw_obj.exit_gw_api()

    def continue_app(self):
        self.gw_init()
        self.gw_reset_and_config()

    def exit_app(self):
        try:
            self.gw_obj.reset_gw()
            time.sleep(1)
            self.gw_obj.exit_gw_api()
            self.resolver_handler.join(10)
            if self.resolver_handler.is_alive():
                self.logger.warning('resolver process is still running')
            if not self.resolved_q.empty():
                self.merge_resolved_data()
        except Exception as e:
            self.logger.warning(f'GetPackets: exit_app: got exception: {e}')
            if not self.exception_q.full():
                self.exception_q.put(f'GetPackets: exit_app: {e}')

    def get_packets_df(self):
        return self.packets_df


if __name__ == '__main__':
    from wiliot_core import set_logger
    RUN_TIME = 60

    tag_pref_logger_path, tag_pref_logger = set_logger(app_name='ReelVerification', dir_name='reel_verification',
                                                       file_name='verification_log')
    stop_event = multiprocessing.Event()
    user_input = {
        'energy_pattern': '51',
        'time_profile': '5,15',
        'ble_power': '22', 'sub1g_power': '29', 'sub1g_freq': '925000',
        'scan_ch': '37',
        'is_listen_bridge': 'yes',
        'owner_id': 'wiliot-ops',
        'env': 'prod'
    }
    v = ReelVerification(user_input=user_input, stop_event=stop_event,
                         logger_name=tag_pref_logger.name, logger_path=tag_pref_logger_path)

    t_i = time.time()
    while time.time() - t_i < RUN_TIME:
        time.sleep(1)
        df = v.get_packets_df()
        print(f'n unique adva: {len(df)}')
    # stop run
    stop_event.set()

    df = v.get_packets_df()
    df.index = df.index.set_names(['adv_address'])
    df = df.reset_index()
    df_path = tag_pref_logger_path.replace('.log', '_packets_df.csv')
    print(f'saving data at: {df_path}')
    df.to_csv(df_path, index=False)

    print('done')



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
from shutil import copyfile
from os import makedirs
import pygubu
from tkinter import Tk, Toplevel
from os.path import join, abspath, dirname, isfile, isdir
from json import dump, load
from copy import deepcopy
from wiliot_core import WiliotDir
from wiliot_testers.utils.get_version import get_version

TESTER_NAME = 'sample'
VER = get_version()
MEASURED = {'button': 'Measured', 'label': '[meter]'}
MANUAL = {'button': 'Manual', 'label': '[db]'}
DEFAULT_EMULATED_SURFACE = 'no simulation'

wiliot_dir = WiliotDir()
wiliot_dir.create_tester_dir(TESTER_NAME)
sample_test_dir = wiliot_dir.get_tester_dir(TESTER_NAME)
OUTPUT_DIR = abspath(join(sample_test_dir, 'logs'))
CONFIGS_DIR = abspath(join(sample_test_dir, 'configs'))
if not isdir(abspath(join(OUTPUT_DIR))):
    makedirs(abspath(join(OUTPUT_DIR)))
if not isdir(abspath(join(CONFIGS_DIR))):
    makedirs(abspath(join(CONFIGS_DIR)))


class ConfigsGui(object):
    is_gui = False
    atten_cur_mode = MANUAL
    test_config = ''
    config = ''
    configs_test = {}
    paramDict = {}
    defaultConfigsDict = {}
    default_surfaces = {}
    all_configs_fields = []
    builder = None
    ttk = None
    mainwindow = None
    
    def __init__(self, top_builder=None, tk_frame=None):
        self.top_builder = top_builder
        self.tk_frame = tk_frame
        self.config_path = abspath(join(dirname(__file__), 'configs', '.default_test_configs.json'))
        if isfile(self.config_path):
            with open(self.config_path, 'r') as jsonFile:
                self.configs_test = load(jsonFile)
                if len(self.configs_test):
                    self.all_configs_fields = list(self.configs_test[next(iter(self.configs_test))].keys())
            self.fix_antenna_type()

        with open(abspath(join(dirname(__file__), 'configs', '.default_surfaces.json')), 'r') as jsonFile:
            self.default_surfaces = load(jsonFile)

        self.copy_config_file()

    def copy_config_file(self):
        copyfile(self.config_path, abspath(join(CONFIGS_DIR, f'.default_test_configs(ViewOnly)_{VER}.json')))

    def gui(self):
        self.builder = builder = pygubu.Builder()
        ui_file = abspath(join(abspath(dirname(__file__)), 'utils', 'configs.ui'))
        self.builder.add_from_file(ui_file)
        
        img_path = abspath(join(abspath(dirname(__file__)), ''))
        builder.add_resource_path(img_path)
        img_path = abspath(join(abspath(dirname(__file__)), 'utils'))
        builder.add_resource_path(img_path)

        if self.tk_frame is not None:
            self.ttk = Toplevel(self.tk_frame)
        else:
            self.ttk = Tk()

        self.ttk.title("Sample Test Configs")
        
        self.mainwindow = self.builder.get_object('mainwindow', self.ttk)
        
        self.builder.connect_callbacks(self)
        
        self.ttk.protocol("WM_DELETE_WINDOW", self.close_cb)
        self.ttk.lift()
        self.ttk.attributes("-topmost", True)
        self.ttk.attributes("-topmost", False)
        
        self.set_gui_defaults()
        
        self.is_gui = True
        self.ttk.mainloop()
    
    def set_gui_defaults(self):
        temp_dict = deepcopy(self.configs_test)
        self.builder.get_object('configsList')['values'] = [key for key, item in temp_dict.items()
                                                            if isinstance(item, dict)]
        if temp_dict.get(self.config):
            self.builder.get_object('configsList').set(self.config)
            self.top_builder.tkvariables.get('testTime').set(temp_dict[self.config]['testTime'])
            self.builder.get_object('EmulateSurface')['values'] = tuple(
                self.default_surfaces['EmulateSurface'].keys())
            self.builder.get_object('EmulateSurface').set(DEFAULT_EMULATED_SURFACE)
            self.builder.get_object('antennaType')['values'] = ['TEO', 'TIKI']
            for param, value in temp_dict[self.config].items():
                if self.builder.tkvariables.get(param) is not None:
                    self.builder.tkvariables.get(param).set(value)
                else:
                    pass
        self.builder.get_object('save')['state'] = 'normal'
    
    def fix_antenna_type(self):
        tempDict = deepcopy(self.configs_test)
        for config, params in tempDict.items():
            if 'antennaType' in params.keys() and params['antennaType']:
                fixedAntenna = 'TIKI' if params['antennaType'].lower() in ['dual', 'tiki'] else 'TEO'
                self.configs_test[config]['antennaType'] = fixedAntenna

    def is_gui_opened(self):
        return self.is_gui
    
    def get_params(self):
        return self.paramDict
    
    def get_configs(self):
        temp_dict = deepcopy(self.configs_test)
        return temp_dict
    
    def set_params(self, test_config):
        if test_config not in self.configs_test.keys():
            config_options = list(self.configs_test.keys())
            test_config = config_options[0]
        self.paramDict = self.configs_test[test_config].copy()
        if 'EmulateSurface' in self.configs_test[test_config].keys():
            surface = self.configs_test[test_config]['EmulateSurface']
            self.paramDict['EmulateSurfaceValue'] = self.configs_test[test_config]['EmulateSurfaceValue'] = \
                self.default_surfaces['EmulateSurface'][surface]
        self.top_builder.tkvariables.get('testTime').set(self.paramDict['testTime'])
    
    def set_default_config(self, test_config):
        self.test_config = test_config
        self.config = test_config

    def config_set(self, config):
        self.config = config
        self.set_params(config)
        if self.is_gui:
            self.builder.get_object('configsList').set(config)
            self.set_gui_defaults()
        # except BaseException:
        #     pass

    # ############## GUI Callbacks  #######################

    def choose_antenna_type_cb(self, *args):
        antenna = self.builder.tkvariables.get('antennaType').get()
        antenna = 'TIKI' if antenna.lower() in ['dual', 'tiki'] else 'TEO'
        self.builder.tkvariables.get('antennaType').set(antenna)

    def reset_cb(self):
        def_dict = deepcopy(self.defaultConfigsDict)
        self.configs_test.update(def_dict)
        self.set_gui_defaults()

    def atten_mode_cb(self):
        self.atten_cur_mode = MEASURED if self.atten_cur_mode == MANUAL else MANUAL
        self.builder.tkvariables.get('atten_mode').set(self.atten_cur_mode['button'])
        ble_label = self.builder.tkvariables.get('attenBleLabel')
        ble_label.set(ble_label.get().split()[0] + ' ' + self.atten_cur_mode['label'])
        lora_label = self.builder.tkvariables.get('attenLoRaLabel')
        lora_label.set(lora_label.get().split()[0] + ' ' + self.atten_cur_mode['label'])

    def close_cb(self):
        self.is_gui = False
        self.ttk.destroy()

    def save_cb(self):
        self.config = config = self.builder.get_object('configsList').get()

        if config not in self.configs_test.keys():
            self.configs_test[config] = {}

        for param in self.all_configs_fields:
            value = self.builder.tkvariables.get(param)
            if value is not None:
                self.configs_test[config][param] = value.get()
        self.builder.get_object('configsList')['values'] = [key for key, item in self.configs_test.items()
                                                            if isinstance(item, dict)]
        self.top_builder.get_object('test_config')['values'] = [key for key, item in self.configs_test.items()
                                                                if isinstance(item, dict)]
        self.top_builder.get_object('test_config').set(config)

        with open(self.config_path, 'w+') as jsonFile:
            dump(self.configs_test, jsonFile, indent=4)
        self.copy_config_file()

        self.set_params(config)
        print(f'{config} configuration saved successfully.')

    def config_select_cb(self, *args):
        self.config = config = self.builder.get_object('configsList').get()
        self.builder.get_object('save')['state'] = 'normal'
        self.top_builder.get_object('test_config').set(config)
        self.reset_cb()



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

from assembly_yield_tester import MainWindow as BaseMainWindow
from assembly_yield_tester import *

INLAY_INFO_COLUMNSPAN = 200
CONVO_TIME_BETWEEN_MATRICES = 0.1
CONVO_RSSI_THRESHOLD = int(user_inputs.get('rssi_threshold', '50') or '50')
TESTER_TYPE = 'conversion_yield_tester'


class MainWindow(BaseMainWindow):

    def __init__(self, ):
        super().__init__()
        self.last_tags_number_of_advas = 0
        self.tester_type = 'conversion-yield'

    def init_processes(self, inlay_select, time_between_matrices=CONVO_TIME_BETWEEN_MATRICES,
                       rssi_th=CONVO_RSSI_THRESHOLD):
        super().init_processes(inlay_select=inlay_select, time_between_matrices=time_between_matrices, rssi_th=rssi_th)

    def get_log_folder_name(self):
        return '_'.join([self.assembled_reel, formatted_today, cur_time_formatted])

    def get_cmn(self, day_str, time_str):
        return '_'.join([self.assembled_reel, day_str, time_str])

    @staticmethod
    def open_session_layout(previous_input, inlay_info):
        open_session_layout = {
            'assembled_reel': {'text': 'Assembled Reel:', 'value': previous_input.get('assembled_reel', ''),
                               'widget_type': 'entry'},
            'rows_num': {'text': 'Number of lanes:', 'value': previous_input.get('rows_num', 1), 'widget_type': 'combobox', 'options': [1, 2, 3]},
            'parts_row': [
                {'part_1': {'text': 'PART 1', 'value': previous_input.get('parts_row_part_1', '')}},
                {'part_2': {'text': 'PART 2', 'value': previous_input.get('parts_row_part_2', '')}},
                {'part_3': {'text': 'PART 3', 'value': previous_input.get('parts_row_part_3', '')}},
            ],
            'inlay_dict': [
                {'inlay': {'text': 'Inlay:', 'value': previous_input.get('inlay_dict_inlay', ''),
                           'widget_type': 'combobox',
                           'options': list(inlays.keys())}},
                {'inlay_info': {'widget_type': 'label', 'value': inlay_info['inlay_info'],
                                'columnspan': INLAY_INFO_COLUMNSPAN}},
            ],
            'tester_station_name': {'text': 'Tester Station:', 'value': previous_input.get('tester_station_name', ''),
                                    'widget_type': 'entry'},
            'comments': {'text': 'Comments:', 'value': previous_input.get('comments', ''), 'widget_type': 'entry'},
            'operator': {'text': 'Operator:', 'value': previous_input.get('operator', ''), 'widget_type': 'entry'},
            'conversion_type': {'text': 'Conversion:', 'value': previous_input.get('conversion_type', ''),
                                'widget_type': 'combobox', 'options': inlay_info["conv_opts"]},
            'surface': {'text': 'Surface:', 'value': previous_input.get('surface', ''), 'widget_type': 'combobox',
                        'options': inlay_info["surfaces"]},
            'window_size': {'text': 'Window Size for Analysis:', 'value': previous_input['window_size'],
                            'widget_type': 'entry'},
            'do_resolve': {'text': 'Get External Id from Cloud', 'value': previous_input['do_resolve']},
            'owner_id': {'text': 'Owner Id for Cloud Connection', 'value': previous_input['owner_id']},

        }

        return open_session_layout

    def open_session(self, cols_or_rows='rows_num'):
        super().open_session(cols_or_rows='rows_num')

    def get_current_plot_title(self):
        yield_type = 'Advas Yield' if self.yield_type == 'Advas' else 'External Ids Yield'

        return f'{yield_type} of last {self.window_size} tags'

    def setup_inlay_parameters(self, values_out, energy_pat, time_pro, thermodes_col=1, rows_num=1,
                               yield_tester_type=TESTER_TYPE):
        self.assembled_reel = values_out.get('assembled_reel', '')
        super().setup_inlay_parameters(values_out, energy_pat, time_pro, thermodes_col, yield_tester_type=TESTER_TYPE,
                                       rows_num=values_out.get('rows_num', ''))
        parts = []

        for i in range(1, int(self.rows_number) + 1):
            part_value = values_out[f'parts_row_part_{i}']
            if part_value:
                parts.append(part_value)
            self.logger.info(f'part{i} is {part_value}')
        self.lane_ids = ','.join(parts)
        self.logger.info(f'lanes_id: {self.lane_ids}')


if __name__ == '__main__':
    m = MainWindow()
    m.run()



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
'''
Eduard Tatievski
'''
import re
import sys
import threading

import matplotlib.pyplot as plt
from wiliot_core import PacketList
from wiliot_tools.utils.wiliot_gui.wiliot_gui import WiliotGui

from wiliot_testers.calibration_test.calibration_test_by_tbp import *
from wiliot_core.packet_data import packet_list, packet, tag_collection
from wiliot_tools.test_equipment.test_equipment import Attenuator, EquipmentError
from wiliot_testers.wiliot_tester_log import *
import os
import pathlib
import datetime
import csv
import pandas as pd
import collections


class Harvester(object):

    def __init__(self):
        # Getting all wanted values
        self.new_bb_list = None
        self.num_of_chambers = 1  # Number of chambers connected
        self.min_tags_answering = 1
        self.gui_values = self.init_gui()
        # Create log folder if not exist and build new name
        self.init_name_path()
        # Create log class
        self.my_logger = WiliotTesterLog(self.run_name_date)
        self.my_logger.set_logger(self.log_path, tester_name='Harvester')
        # Connect to GW and Attenuator if selected
        self.init_hw()
        # Build new power list according to gui inputs
        # self.power_range = self.build_range()
        self.power = re.split('(d)', self.gui_values['power'])[0]
        self.my_logger.results_logger.info('Starting scan on power: {}'.format(self.power))
        self.frequency_range = list(range(int(self.gui_values['start_freq']), int(self.gui_values['stop_freq']) + 1,
                                          int(self.gui_values['step'])))
        self.all_results = pd.DataFrame({'Adva': [], 'Frequency': [], 'TBP': [], 'BB': []})
        self.scan()

    def calibrate_bb(self):
        if self.gw_obj.connected:
            self.gw_obj.reset_buffer()
            self.gw_obj.start_continuous_listener()
        for bb_val in reversed(self.bb_list):
            self.gw_obj.config_gw(energy_pattern_val=self.gui_values['pattern'],
                                  time_profile_val=[self.gui_values['onTime'], self.gui_values['totalTime']],
                                  output_power_val='pos3dBm',
                                  beacons_backoff_val=bb_val)

    def build_range(self):
        if not self.sub1g_enable:
            if self.attn_enable:
                for index, nrg in enumerate(self.abs_att_power_list):
                    if self.gui_values['lowerPower'] == nrg:
                        low_limit = index
                    if self.gui_values['higherPower'] == nrg:
                        high_limit = index
                power_range = self.abs_att_power_list[low_limit:high_limit + 1]
            else:
                for index, nrg in enumerate(self.abs_power_list):
                    if self.gui_values['lowerPower'] == nrg:
                        low_limit = index
                    if self.gui_values['higherPower'] == nrg:
                        high_limit = index
                power_range = self.abs_power_list[low_limit:high_limit + 1]

        else:
            if self.attn_enable:
                for index, nrg in enumerate(self.valid_attn_sub1g_output_power):
                    if self.gui_values['lowerPower'] == nrg:
                        low_limit = index
                    if self.gui_values['higherPower'] == nrg:
                        high_limit = index
                power_range = self.valid_attn_sub1g_output_power[low_limit:high_limit + 1]
            else:
                for index, nrg in enumerate(self.valid_sub1g_output_power):
                    if self.gui_values['lowerPower'] == nrg:
                        low_limit = index
                    if self.gui_values['higherPower'] == nrg:
                        high_limit = index
                power_range = self.valid_sub1g_output_power[low_limit:high_limit + 1]

        output_power = []
        for power in power_range:
            if int(power[:-3]) < 0:
                output_power.append('neg' + power[1:])
            else:
                output_power.append('pos' + power)

        return output_power

    def convert_power(self, power, sub1g=False):
        gw_power = None
        attn_power = None
        # Convert from neg3dBm to -3
        if power[:3] == 'neg':
            abs_power = 0 - int(power[3:-3])
        elif power[:3] == 'pos':
            abs_power = int(power[3:-3])
        else:
            abs_power = int(power[:-3])

        # Find the value for attenuator if available
        if sub1g:
            if self.attn_enable:
                for power_scan in self.valid_sub1g_output_power:
                    if int(power_scan[:-3]) > abs_power:
                        gw_power = int(power_scan[:-3])
                        attn_power = int(gw_power[:-3]) - abs_power
            else:
                gw_power = abs_power
                attn_power = 0

        else:
            if self.attn_enable:
                for power_scan in self.abs_power_list:
                    if int(power_scan[:-3]) > abs_power:
                        gw_power = int(power_scan[:-3])
                        attn_power = int(gw_power[:-3]) - abs_power
            else:
                gw_power = abs_power
                attn_power = 0
        return gw_power, attn_power

    def init_hw(self):

        try:
            self.gw_obj = WiliotGateway(auto_connect=True, logger_name=self.my_logger.gw_logger.name, verbose=False)
            time.sleep(1)
            if self.gw_obj.connected:
                self.gw_obj.reset_gw()
                self.gw_obj.reset_buffer()
                time.sleep(1.5)
                self.gw_obj.start_continuous_listener()
                self.gw_obj.write('!set_tester_mode 1', with_ack=True)
                self.gw_obj.write('!pl_gw_config 1', with_ack=True)
                self.gw_obj.write('!enable_brg_mgmt 0', with_ack=True)
                self.gw_obj.write('!listen_to_tag_only 1', with_ack=True)
                self.gw_obj.write('!energizing_prob_set 100', with_ack=True)
                self.my_logger.gw_logger.info('GW config finished')
            else:
                raise EquipmentError('Gateway Error - Verify WiliotGateway connection')
        except Exception as e:
            self.my_logger.gw_logger.warning(e)
            raise EquipmentError('Gateway Error - Verify WiliotGateway connection')
            self.gw_obj.close_port()

        if self.attn_enable:
            self.attntype = self.gui_values['attntype']
            self.attnval = int(self.gui_values['attnval'])

            try:
                self.attn_obj = Attenuator(str(self.attntype)).GetActiveTE()
                current_attn = self.attn_obj.Getattn()
                self.my_logger.gw_logger.info('Current attenuator values: {}'.format(current_attn))
            except Exception:
                raise EquipmentError('Attenuator Error - Verify Attenuator connection')

    def init_name_path(self):
        self.log_path = os.path.join('', 'logs')
        if not os.path.exists(self.log_path):
            pathlib.Path(self.log_path).mkdir(parents=True, exist_ok=True)
        self.time_start = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        self.run_name_date = str(self.gui_values['testName']) + '_' + str(self.time_start)
        self.full_name_path = os.path.join(self.log_path, self.run_name_date)
        self.packets_csv_path = os.path.join(self.full_name_path, self.run_name_date + '.csv')
        self.test_csv_path = os.path.join(self.full_name_path, self.run_name_date + '_tests.csv')

    def init_gui(self):

        # self.bb_list = [0, 2, 7, 12, 19, 20, 21, 22, 23, 24, 25, 27, 29, 30, 31, 32, 33, 36, 40]
        self.bb_list = [40, 36, 33, 32, 31, 30, 29, 27, 25, 24, 23, 22, 21, 20, 19, 12, 7, 2, 0]
        # self.bb_list = [12, 7, 2, 0]

        self.valid_sub1g_output_power = ['17dBm', '18dBm', '19dBm', '20dBm', '21dBm', '22dBm',
                                         '23dBm', '24dBm', '25dBm', '26dBm', '27dBm', '28dBm', '29dBm']

        self.valid_attn_sub1g_output_power = ['-10dBm', '-9dBm', '-8dBm', '-7dBm', '-6dBm', '-5dBm',
                                              '-4dBm', '-3dBm', '-2dBm', '-1dBm', '0dBm', '1dBm',
                                              '2dBm', '3dBm', '4dBm', '5dBm', '6dBm', '7dBm', '8dBm',
                                              '9dBm', '10dBm', '11dBm', '12dBm', '13dBm', '14dBm', '15dBm',
                                              '16dBm', '17dBm', '18dBm', '19dBm', '20dBm', '21dBm', '22dBm',
                                              '23dBm', '24dBm']

        self.abs_power_list = ['-21dBm', '-18dBm', '-15dBm', '-12dBm', '-8dBm', '-5dBm', '-2dBm',
                               '-1dBm', '0dBm', '1dBm', '2dBm', '3dBm', '4dBm', '6dBm',
                               '10dBm', '15dBm', '20dBm', '21dBm', '22dBm']

        self.abs_att_power_list = ['-35dbm', '-34dbm', '-33dbm', '-32dbm', '-31dbm', '-30dbm',
                                   '-29dbm', '-28dbm', '-27dbm', '-26dbm', '-25dbm', '-24dbm',
                                   '-23dbm', '-22dbm', '-21dbm', '-20dbm', '-19dbm', '-18dbm',
                                   '-17dbm', '-16dbm', '-15dbm', '-14dbm', '-13dbm', '-12dbm',
                                   '-11dbm', '-10dbm', '-9dbm', '-8dbm', '-7dbm', '-6dbm',
                                   '-5dbm', '-4dbm', '-3dbm', '-2dbm', '-1dbm', '0dbm',
                                   '1dbm', '2dbm', '3dbm', '4dbm', '5dbm', '6dbm',
                                   '7dbm', '8dbm', '9dbm', '10dbm', '11dbm',
                                   '12dbm', '13dbm', '14dbm', '15dbm', '16dbm', '17dbm']

        self.valid_output_power_vals = [
            {'abs_power': -21, 'gw_output_power': 'neg20dBm', 'bypass_pa': '1'},
            {'abs_power': -18, 'gw_output_power': 'neg16dBm', 'bypass_pa': '1'},
            {'abs_power': -15, 'gw_output_power': 'neg12dBm', 'bypass_pa': '1'},
            {'abs_power': -12, 'gw_output_power': 'neg8dBm', 'bypass_pa': '1'},
            {'abs_power': -8, 'gw_output_power': 'neg4dBm', 'bypass_pa': '1'},
            {'abs_power': -5, 'gw_output_power': 'pos0dBm', 'bypass_pa': '1'},
            {'abs_power': -2, 'gw_output_power': 'pos2dBm', 'bypass_pa': '1'},
            {'abs_power': -1, 'gw_output_power': 'pos3dBm', 'bypass_pa': '1'},
            {'abs_power': 0, 'gw_output_power': 'pos4dBm', 'bypass_pa': '1'},
            {'abs_power': 1, 'gw_output_power': 'pos5dBm', 'bypass_pa': '1'},
            {'abs_power': 2, 'gw_output_power': 'pos6dBm', 'bypass_pa': '1'},
            {'abs_power': 3, 'gw_output_power': 'pos7dBm', 'bypass_pa': '1'},
            {'abs_power': 4, 'gw_output_power': 'pos8dBm', 'bypass_pa': '1'},
            {'abs_power': 6, 'gw_output_power': 'neg12dBm', 'bypass_pa': '0'},
            {'abs_power': 10, 'gw_output_power': 'neg8dBm', 'bypass_pa': '0'},
            {'abs_power': 15, 'gw_output_power': 'neg4dBm', 'bypass_pa': '0'},
            {'abs_power': 20, 'gw_output_power': 'pos0dBm', 'bypass_pa': '0'},
            {'abs_power': 21, 'gw_output_power': 'pos2dBm', 'bypass_pa': '0'},
            {'abs_power': 22, 'gw_output_power': 'pos3dBm', 'bypass_pa': '0'}]

        params_dict = {
            'testName': {
                'text': 'Test Name',
                'value': '',
                'widget_type': 'entry',
            },
            'tagsNum': {
                'text': 'Num of tags',
                'value': self.min_tags_answering,
                'widget_type': 'entry',
            },
            'testTime': {
                'text': 'Test Time',
                'value': '10',
                'widget_type': 'entry',
            },
            'start_freq': {
                'text': 'Start [MHz]',
                'value': '2402',
                'widget_type': 'entry',
                'group': 'Frequency range'
            },
            'stop_freq': {
                'text': 'Stop [MHz]',
                'value': '2480',
                'widget_type': 'entry',
                'group': 'Frequency range'
            },
            'step': {
                'text': 'Step [MHz]',
                'value': '2',
                'widget_type': 'entry',
                'group': 'Frequency range'
            },
            'power': {
                'text': 'Energy',
                'value': self.abs_power_list[0],
                'options': self.abs_power_list,
                'widget_type': 'combobox',
            },
            'lowerBB': {
                'text': 'Beacons backoff (lower)',
                'value': self.bb_list[0],
                'options': self.bb_list,
                'widget_type': 'combobox',
            },
            'higherBB': {
                'text': 'Beacons backoff (higher)',
                'value': self.bb_list[-1],
                'options': self.bb_list,
                'widget_type': 'combobox',
            },
            '2.4mode': {
                'text': '2.4Ghz Mode',
                'value': '2.4Ghz',
                'widget_type': 'button',
            },
            'sub1gmode': {
                'text': 'Sub1G Mode',
                'value': 'Sub1G',
                'widget_type': 'button',
            },
            'onTime': {
                'text': 'Time profile - On',
                'value': '5',
                'widget_type': 'entry',
            },
            'totalTime': {
                'text': 'Off',
                'value': '15',
                'widget_type': 'entry',
            },
            'attn': {
                'text': 'Enable Attenuator',
                'value': 'Enable Attenuator',
                'widget_type': 'button',
            },
            '-ATTN-': {
                'text': 'Attenuator Disabled',
                'value': 'Attenuator Disabled',
                'widget_type': 'label',
                'options': {'fg': 'Dark Red'}
            },
            'attntype': {
                'text': 'Attenuator type',
                'value': 'API',
                'options': ('API', 'MCDI', 'MCDI-USB'),
                'widget_type': 'combobox',
            },
            'attntypeen': {
                'text': 'Attenuator type: ',
                'value': '',
                'widget_type': 'label',
            },
            'attnval': {
                'text': 'Attenuation in dBm',
                'value': '0',
                'widget_type': 'entry',
            },
            'attnvalen': {
                'text': 'Attenuation in dBm',
                'value': '',
                'widget_type': 'label',
            },
            '-OUTPUT-': {
                'text': '',
                'value': '',
                'widget_type': 'label',
                'options': {'fg': 'red', 'font': 'bold'},
            }
        }

        wg = WiliotGui(params_dict=params_dict, do_button_config=False)
        self.attn_enable = False
        self.sub1g_enable = False

        def on_submit_button():
            test_name = wg.widgets.get('testName').get()
            start_freq = int(wg.widgets.get('start_freq').get())
            stop_freq = int(wg.widgets.get('stop_freq').get())
            if ' ' in test_name or '/' in test_name or '\\' in test_name or test_name == '':
                wg.update_widget(widget_key='-OUTPUT-', new_value='Please enter valid run name without / or spaces',
                                 color='red')
            if self.sub1g_enable:
                if start_freq not in range(849, 941) or stop_freq not in range(849, 941):
                    wg.update_widget(widget_key='-OUTPUT-', new_value='Please enter valid Frequency range for Sub1G',
                                     color='red')
            else:
                if start_freq not in range(2400, 2490) or stop_freq not in range(2400, 2490):
                    wg.update_widget(widget_key='-OUTPUT-', new_value='Please enter valid Frequency range for 2.4GHz',
                                     color='red')
            self.min_tags_answering = int(wg.widgets.get('tagsNum').get())
            start_index = int(wg.widgets.get('lowerBB').get())
            higher_bb = int(wg.widgets.get('higherBB').get())
            end_index = self.bb_list.index(higher_bb)
            direction = -1 if start_index > end_index else 1
            self.new_bb_list = self.bb_list[start_index:end_index + direction:direction]
            wg.on_submit()

        def on_sub1gmode_click():

            self.sub1g_enable = True

            wg.update_widget(widget_key='2.4mode', new_value='2.4Ghz', color='#821515')
            wg.update_widget(widget_key='sub1gmode', new_value='Sub1G', color='#158225')
            wg.update_widget(widget_key='power', new_value=self.valid_sub1g_output_power[0], options=tuple(self.valid_sub1g_output_power))
            wg.update_widget(widget_key='start_freq', new_value='850')
            wg.update_widget(widget_key='stop_freq', new_value='940')

        def on_24mode_click():

            self.sub1g_enable = False

            wg.update_widget(widget_key='2.4mode', new_value='2.4Ghz', color='#158225')
            wg.update_widget(widget_key='sub1gmode', new_value='Sub1G', color='#821515')
            wg.update_widget(widget_key='power', new_value=self.abs_power_list[0], options=tuple(self.abs_power_list))

            # Update the 'start_freq' and 'stop_freq' entries
            wg.update_widget(widget_key='start_freq', new_value='2402')
            wg.update_widget(widget_key='stop_freq', new_value='2480')

        def on_attn_click():
            self.attn_enable = not self.attn_enable

            if self.attn_enable:
                wg.update_widget(widget_key='attn', new_value='Disable Attenuator', color='#821515')
                wg.update_widget(widget_key='-ATTN-', new_value='Attenuator Enabled', color='Dark Green')

                wg.update_widget(widget_key='attntype', disabled=False)
                wg.update_widget(widget_key='attntypeen', disabled=False)
                wg.update_widget(widget_key='attnval', disabled=False)
                wg.update_widget(widget_key='attnvalen', disabled=False)
            else:
                wg.update_widget(widget_key='attn', new_value='Enable Attenuator', color='#158225')
                wg.update_widget(widget_key='-ATTN-', new_value='Attenuator Disabled', color='Dark Red')

                wg.update_widget(widget_key='attntype', disabled=True)
                wg.update_widget(widget_key='attntypeen', disabled=True)
                wg.update_widget(widget_key='attnval', disabled=True)
                wg.update_widget(widget_key='attnvalen', disabled=True)

        wg.button_configs(submit_command=on_submit_button)
        wg.add_event(widget_key='sub1gmode', command=on_sub1gmode_click, event_type='button')
        wg.add_event(widget_key='2.4mode', command=on_24mode_click, event_type='button')
        wg.add_event(widget_key='attn', command=on_attn_click, event_type='button')
        on_attn_click()
        values = wg.run()
        return values

    def packet_filter(self, packet_list, filter_param='rssi', param_limits=None):
        """
        filter the packets from the packet_list according to the filter_param and the param_limits.
        packets that were filtered out of the test does no take into account during the test and considered as noise
        :param packet_list: the incoming packets list
        :type packet_list: PacketList
        :param filter_param: one of the packet dataframe column names
        :type filter_param: str
        :param param_limits: list of two elements [x, y] that defines the range of the valid param (x,y including)
        :type param_limits: list
        :return: the filtered packet list
        :rtype: PacketList
        """

        def update_custom_data(packet, key, value):
            if key in packet.custom_data.keys():
                packet.custom_data[key].append(value)
            else:
                packet.custom_data[key] = [value]

        if param_limits is None:
            param_limits = [0, float('inf')]
        filtered_packet_list = PacketList()

        for p in packet_list:
            # filter packets from black list:
            if p.packet_data['adv_address'] in self.black_list:
                p.add_custom_data(custom_data={'packet_status': 'blacklist'})
            else:
                # filter packets by filter param:
                packet_param_list = p.extract_packet_data_by_name(filter_param)
                sprinkler_ids = []
                for i, param in enumerate(packet_param_list):
                    if param_limits[0] <= param <= param_limits[-1]:
                        update_custom_data(p, 'packet_status', 'good')
                        sprinkler_ids.append(i)
                    else:
                        update_custom_data(p, 'packet_status', filter_param)

                if len(sprinkler_ids) > 0:
                    filtered_packet_list.append(p.filter_by_sprinkler_id(sprinkler_ids=sprinkler_ids))

        return filtered_packet_list

    def init_timer(self):
        self.timeout = False
        self.success = False
        self.clean_gw()
        self.timer = threading.Timer(int(self.gui_values['testTime']), self.time_expire)

    def start_analyze(self, power=None, freq=None, bb=None):
        tbp = None
        for tag_adva in self.all_tags.tags:
            if self.is_gw_fetal_error(self.all_tags.tags[tag_adva]):
                self.my_logger.gw_logger.warning('gw fetal error was detected')
                self.gw_reset_and_config()
            for packet in self.all_tags.tags[tag_adva].packet_list:
                self.all_tests.append(packet)
            try:
                tbp = self.all_tags.get_avg_tbp_by_id(tag_adva)
                packets_sum = self.all_tags.get_statistics_by_id(tag_adva)['num_packets']
            except Exception as e:
                self.my_logger.logger.info(
                    'Exception receiving TBP or Packets num on tag {}, exception {}'.format(tag_adva, e))
                tbp = None
            new_row = {'Adva': tag_adva, 'Frequency': freq, 'TBP': tbp, 'BB': bb}
            new_df = pd.DataFrame(new_row, index=[0])
            self.all_results = pd.concat([self.all_results, new_df], ignore_index=True)
            details = {'Test Num': self.test_num, 'BB': bb, 'Frequency': freq, 'Power': power, 'Packets': packets_sum,
                       'TBP': tbp}
            if self.all_tests.tags[tag_adva].list_custom_data.__len__() == 0:
                self.all_tests.tags[tag_adva].list_custom_data['Data'] = []
            self.all_tests.tags[tag_adva].list_custom_data['Data'].append(details)

    def get_packets(self, filter_param='rssi', filter_value=100, power=None, freq=None, bb=None):
        """
        receive packets from the gw (via the listener) and filter them according to packet_filter()
        :param filter_param: one of the packet dataframe column names (for packet_filter)
        :type filter_param: str
        :param filter_value: the max valid value of the specified param
        :type filter_value: int
        :return: True if packet were received
        :rtype: bool
        """
        packets_received_list = None
        self.num_good_tags = 0
        self.num_bad_tags = 0
        if self.gw_obj.is_data_available():
            packets_received_list = self.gw_obj.get_packets(action_type=ActionType.ALL_SAMPLE,
                                                            data_type=DataType.PACKET_LIST)
        if packets_received_list:
            for packet in packets_received_list:
                packet_param_list = int(packet.get_average_rssi())
                if filter_value >= packet_param_list:
                    self.my_logger.logger.info('got packet {}'.format(packet.packet_data['raw_packet']))
                    self.all_tags.append(packet=packet.copy())
                else:
                    self.my_logger.logger.info(
                        'got packet {} but with high rssi {}'.format(packet.packet_data['raw_packet'],
                                                                     packet_param_list))
                    continue

                if self.all_tags.tags.__len__() > 0:
                    for packet_list in self.all_tags.tags:
                        try:
                            avg_tbp = self.all_tags.tags[packet_list].get_avg_tbp()
                            if avg_tbp > 0:
                                self.num_good_tags += 1
                        except Exception:
                            pass
                        if self.num_good_tags >= self.min_tags_answering:
                            self.num_bad_tags = self.all_tags.tags.__len__() - self.num_good_tags
                            self.success = True
                        else:
                            continue
                if self.success:
                    self.my_logger.logger.info(
                        'Found {} tags\n{} tags TBP was calculated\n{} tags TBP wasnt calculated\nStarting analyzing'.format(
                            self.all_tags.tags.__len__(), self.num_good_tags, self.num_bad_tags))
                    break

        if self.success:
            self.timer.cancel()
            self.gw_obj.stop_gw_app()
            self.clean_gw()
            self.gw_obj.stop_continuous_listener()
            self.start_analyze(power, freq, bb)

    def is_gw_fetal_error(self, packet_list):
        """
        check if the gw reset itself for some reason
        :param packet_list: the received packets list
        :type packet_list: PacketList
        :return:
        :rtype:
        """
        counter = 0
        for p in packet_list:
            if p.gw_data['stat_param'].size > 1:
                for s in p.gw_data['stat_param']:
                    if s == 0:
                        counter = counter + 1
                        if counter > 1:
                            return True
            else:
                if p.gw_data['stat_param'] == 0:
                    counter = counter + 1
                    if counter > 1:
                        return True
        return False

    def get_result(self):
        return self.all_tests

    def time_expire(self):
        self.timeout = True
        self.timer.cancel()
        # self.gw_obj.stop_gw_app()
        self.my_logger.results_logger.info(
            'Timeout - found {} tags but didnt received enough packets to calculate TBP for all, calculated only for {} tags'.format(
                self.all_tags.tags.__len__(), self.num_good_tags))

    def gw_reset_and_config(self):
        self.gw_obj.reset_gw()
        time.sleep(2)
        self.gw_obj.write('!set_tester_mode 1', with_ack=True)
        self.gw_obj.write('!enable_brg_mgmt 0', with_ack=True)
        self.gw_obj.write('!listen_to_tag_only 1', with_ack=True)
        self.gw_obj.write('!energizing_prob_set 100', with_ack=True)

    def clean_gw(self):
        self.gw_obj.reset_buffer()
        self.gw_obj.reset_listener()

    def data_read_analyze(self, dict=None):

        # Get unique BB values
        self.all_results = self.all_results.dropna(subset=['TBP'])

        unique_bbs = self.all_results['BB'].unique()

        # Set up subplots
        fig, axes = plt.subplots(nrows=1, ncols=unique_bbs.size, figsize=(10, 4))

        # Loop over unique BB values and plot graphs
        for i, bb in enumerate(unique_bbs):
            # Get data for current BB value
            bb_df = self.all_results[self.all_results['BB'] == bb]
            # Get unique Adva names for current BB value
            unique_adv = bb_df['Adva'].unique()

            # Plot graph for current BB value on the appropriate subplot
            ax = axes[i]
            ax.set_title(f"Beacon Backoff = {bb}")
            freq_ticks = sorted(self.all_results['Frequency'].unique())
            ax.set_xticks(self.x_grid, rotation=90)
            ax.set_xlabel("Frequency")
            ax.set_ylabel("TBP")
            for adv in unique_adv:
                adv_df = bb_df[bb_df['Adva'] == adv]
                ax.plot(adv_df['Frequency'], adv_df['TBP'], label=adv)
            ax.legend()

        # Show the plot
        plt.show()

    def finish(self):
        # Start writing all data to file
        self.all_results = self.all_results.groupby('Adva').apply(
            lambda x: x.sort_values(by=['Frequency', 'BB'], ascending=[True, False])).reset_index(drop=True)

        self.my_logger.results_logger.info(self.all_results)
        for i in self.all_results:
            self.my_logger.results_logger.info(i)
        self.all_packet_list = self.all_tests.to_packet_list()
        all_data = []
        custom_data = {'Test Num': '', 'Adva': '', 'BeaconBackoff': '', 'Frequency': '', 'Power': '', 'Packets': '',
                       'TBP': ''}
        for packet_list in self.all_tests.tags:
            for custom in self.all_tests.tags[packet_list].list_custom_data['Data']:
                p_list_data = {'Test Num': '', 'Adva': '', 'BeaconBackoff': '', 'Frequency': '', 'Power': '',
                               'Packets': '', 'TBP': ''}
                p_list_data['Adva'] = packet_list
                p_list_data['Test Num'] = custom['Test Num']
                p_list_data['BeaconBackoff'] = custom['BB']
                p_list_data['Frequency'] = custom['Frequency']
                p_list_data['Power'] = custom['Power']
                p_list_data['Packets'] = custom['Packets']
                p_list_data['TBP'] = custom['TBP'] if custom['TBP'] is not None else 0
                all_data.append(p_list_data)
        try:
            with open(self.test_csv_path, 'w', newline='') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=custom_data)
                writer.writeheader()
                writer.writerows(all_data)

        except IOError:
            print("I/O error")

        try:
            self.all_tests.to_csv(path=self.packets_csv_path, val='adv_address')
        except Exception:
            self.my_logger.logger.warning('Empty list')
        # Read data from file organized with pandas

        # Draw graphs according to post process correlated to TBP requested
        if len(self.frequency_range) > 8:
            self.x_grid = self.frequency_range[::3]
            if len(self.frequency_range) > 15:
                self.x_grid = self.frequency_range[::4]
                if len(self.frequency_range) > 28:
                    self.x_grid = self.frequency_range[::5]
        else:
            self.x_grid = self.frequency_range
        self.data_read_analyze()
        # The End
        self.my_logger.logger.info('Finished')

    def config_gw(self, power, freq, bb):
        if not self.sub1g_enable:
            # ---------------------------------------------------------------------------------------------------------------------
            # 2.4 + Attenuator
            if self.attn_enable:
                # abs_power_val, abs_attn_val = self.convert_power(power, sub1g=False)
                # for list_power_val in self.valid_output_power_vals:
                #     if abs_power_val == list_power_val['abs_power']:
                #         gw_out_power = list_power_val['gw_output_power']
                self.my_logger.logger.info(
                    'Setting GW power: {}, Attenuation value: {} Frequency {} Beacaon Backoff val {}'.format(
                        power, self.attnval, freq, bb))
                self.attn_obj.Setattn(self.attnval)
                self.gw_obj.config_gw(energy_pattern_val=18,
                                      time_profile_val=[self.gui_values['onTime'],
                                                        self.gui_values['totalTime']],
                                      beacons_backoff_val=bb,
                                      received_channel='37',
                                      effective_output_power_val=int(power),
                                      start_gw_app=True,
                                      with_ack=True)
            # ---------------------------------------------------------------------------------------------------------------------
            # 2.4 + No Attenuator
            else:
                # abs_power_val, abs_attn_val = self.convert_power(power, sub1g=False)
                # if int(abs_power_val) > 5:
                #     bypass_pa_val = 0
                # else:
                #     bypass_pa_val = 1
                # for list_power_val in self.valid_output_power_vals:
                #     if abs_power_val == int(list_power_val['abs_power']):
                #         gw_out_power = list_power_val['gw_output_power']
                self.my_logger.logger.info(
                    'Setting GW power: {}, No attenuation Frequency {} Beacaon Backoff val {}'.format(
                        power, freq, bb))
                self.gw_obj.config_gw(energy_pattern_val=18,
                                      time_profile_val=[self.gui_values['onTime'],
                                                        self.gui_values['totalTime']],
                                      beacons_backoff_val=bb,
                                      received_channel='37',
                                      effective_output_power_val=int(power),
                                      start_gw_app=True,
                                      with_ack=True)
        # ---------------------------------------------------------------------------------------------------------------------
        # Sub1G
        else:
            # ---------------------------------------------------------------------------------------------------------------------
            # Sub1G + Attenuator
            if self.attn_enable:
                # abs_power_val, abs_attn_val = self.convert_power(power, sub1g=True)
                self.my_logger.logger.info(
                    'Setting Sub1G GW power: {}, Attenuation value: {} Frequency {} Beacaon Backoff val {}'.format(
                        power, self.attnval, freq, bb))
                self.attn_obj.Setattn(self.attnval)
                self.gw_obj.config_gw(energy_pattern_val=50,
                                      time_profile_val=[self.gui_values['onTime'],
                                                        self.gui_values['totalTime']],
                                      beacons_backoff_val=bb,
                                      received_channel='37',
                                      effective_output_power_val=22,
                                      sub1g_output_power_val=power,
                                      start_gw_app=True,
                                      with_ack=True)
            # ---------------------------------------------------------------------------------------------------------------------
            # Sub1G + No Attenuator
            else:
                # abs_power_val, abs_attn_val = self.convert_power(power, sub1g=True)
                self.my_logger.logger.info(
                    'Setting Sub1G GW power: {}, No attenuation value Frequency {} Beacaon Backoff val {}'.format(
                        power, freq, bb))
                self.gw_obj.config_gw(energy_pattern_val=50,
                                      time_profile_val=[self.gui_values['onTime'],
                                                        self.gui_values['totalTime']],
                                      beacons_backoff_val=bb,
                                      received_channel='37',
                                      effective_output_power_val=22,
                                      sub1g_output_power_val=power,
                                      start_gw_app=True,
                                      with_ack=True)
        # ---------------------------------------------------------------------------------------------------------------------
        if self.sub1g_enable:
            self.gw_obj.write('!set_dyn_energizing_pattern 6 1 0 0', with_ack=True)
            self.gw_obj.write('!set_beacons_pattern 550 3 2402 2426 2480', with_ack=True)
            self.gw_obj.write('!set_sub_1_ghz_energizing_frequency {}'.format(freq), with_ack=True)

        else:
            self.gw_obj.write('!set_dyn_energizing_pattern 6 0 1 {}'.format(freq), with_ack=True)
            self.gw_obj.write('!set_beacons_pattern 550 3 2402 2426 2480', with_ack=True)

    def scan(self):
        self.all_tests = TagCollection()
        self.test_num = 0
        self.results = []
        # Scan power
        power = self.power
        self.my_logger.logger.info('Sweeping freq {}'.format(self.frequency_range))
        # Scan frequency
        for freq in self.frequency_range:
            # Scan beacon backoff
            for index, bb in enumerate(self.new_bb_list):
                self.test_num += 1
                # 2.4Ghz
                self.all_tags = TagCollection()
                self.gw_obj.start_continuous_listener()
                self.init_timer()  # Init timeout flag and reset success flag
                self.timer.start()  # Start timer thread
                self.my_logger.results_logger.info(
                    '-' * 30 + 'New Test - Power {} Frequency {} Beacon Backoff {}'.format(power, freq,
                                                                                           bb) + '-' * 30)

                self.config_gw(power=power, freq=freq, bb=bb)

                # Start receiving packets
                while not self.timeout and not self.success:  # While function to work until timeout or success test
                    time.sleep(0.01)
                    self.get_packets(power=power, freq=freq,
                                     bb=bb)  # Will start to get packets, then analyze them and add to self.all_tests with custom data of which test it came from

                self.clean_gw()  # clear the buffers
                time.sleep(0.5)

        self.gw_obj.close_port(is_reset=True)
        self.gw_obj.stop_continuous_listener()
        self.finish()


if __name__ == '__main__':
    harvest = Harvester()
    results = harvest.get_result()
    pass



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import os.path
from setuptools_scm import get_version as scm_get_version
import sys
import re


def get_version(action_type="current"):
    """

    @param action_type: can be current - egt the current version, next_patch, next_minor, next_major
    @type action_type: str
    @return: the version number
    @rtype: str
    """
    version_root = os.path.dirname(os.path.dirname(__file__))
    git_root = os.path.dirname(version_root)
    if os.path.isdir(os.path.join(git_root, ".git")):
        version = scm_get_version(root=git_root,
                                  git_describe_command="git describe --long --dirty --tags --match [0-9]*.[0-9]*.[0-9]*")
    elif os.path.isfile(os.path.join(version_root, "version.py")):
        from wiliot_testers.version import __version__
        version = __version__
    else:
        print("Couldn't get version, retrying setupscm get version")
        # give SCM another shot.. shouldn't get here:
        version = scm_get_version(root=git_root,
                                  git_describe_command="git describe --long --dirty --tags --match [0-9]*.[0-9]*.[0-9]*")

    if action_type == "current":
        return version
    else:
        ver_arr = list(re.split(r"\.", version, 4))
        if len(ver_arr) < 3:
            raise ValueError(f"not enough parts at version {version}!!!")
        if action_type == "next_patch":
            if len(ver_arr) == 3:  # we have a valid patch version and we want to progress by 1:
                ver_arr[2] = int(ver_arr[2]) + 1
        elif action_type == "next_minor":
            ver_arr[2] = 0
            ver_arr[1] = int(ver_arr[1]) + 1
        elif action_type == "next_major":
            ver_arr[2] = 0
            ver_arr[1] = 0
            ver_arr[0] = int(ver_arr[0]) + 1
        return f"{ver_arr[0]}.{ver_arr[1]}.{ver_arr[2]}"


if __name__ == '__main__':
    num_args = len(sys.argv)
    if num_args == 1:
        type = 'current'
    else:
        type = sys.argv[1]

    __version__ = get_version(type)
    print(__version__)



"""
  Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

lang_map = {
            'English': 'En',
            'Russian': 'Ru',
            'Hebrew': 'He'
        }

moh_instruction = {
    "Missing Label": {
        "En": [
            "* Please check if there is a missing label case, if not:",
            "    * disconnect and reconnect Arduino",
            "    * check R2R Controller",
            "    * press continue",
            "* Please add a faulty tag inside of the missing label",
            "* Make sure the label on the test site is centered, and click on continue"
        ],
        "Ru": [
            "* Пожалуйста, проверьте, отсутствует ли этикетка",
            "* если нет:",
            "*      Отключите и снова подключите контроллер - Arduino",
            "*      Проверьте настройки контроллера и машины:",
            "*      Нажмите продолжить",
            "* если да:",
            "*      Пожалуйста, добавьте дефектнаю этикетку вместо пустого",
            "*      Убедитесь, что этикетка на тестовом сайте расположена по центру, затем нажмите Продолжить"
        ],
        "He": [
            "*אנא בדוק אם חסרה תווית ",
            "במידה ולא:",
            "נתק וחבר מחדש את הבקר - ארדוינו",
            "בדוק את הבקר והגדרות המכונה",
            "במידה וכן:",
            "*הוסף תווית פגומה במקום שבו חסר",
            "* ודא שהתווית על מקום הבדיקה ממורכזת",
            "*לחץ על המשך"
        ]
    },
    "stop Criteria": {
        "En": [
            "* Please make sure reel is above coupler",
            "* Check antennas connections",
            "* Check test configurations"
        ],
        "Ru": [
            "* Пожалуйста, убедитесь, что лента находится над тестовой антенной",
            "* проверьте подключение антенны",
            "* Проверьте правильные настройки"
        ],
        "He": [
            "* אנא וודא כי הסרט נמצא מעל האנטנת בדיקה",
            "*                     בדוק חיבורי אנטנות",
            "*                     בדוק תקינות הגדרות"
        ]
    },
    "Validation Failed": {
        "En": [
            "* Please check printed label",
            "* If label is correct, please take a picture and send to Wiliot team",
            "* If label is incorrect (different external, black square instead of external id and vice versa):",
            "    * please check the step size in the R2R machine",
            "    * please check the the R2R controller for any errors",
            "    * please press on STOP",
            "* Replace label with a faulty label",
            "* Clean printer and scanner",
            "* Restart printer and scanner if available",
            "* Click on continue after printer and scanner are alive"
        ],
        "Ru": [
            "* Пожалуйста, проверьте напечатанную этикетку",
            "* Если этикетка правильная, сделайте фото и отправьте к Wiliot",
            "* Если есть ошибка несоответствия (ожидается одна этикетку, но проверяется другая этикетка): ",
            "       * проверьте размер шага на машине R2R",
            "       * Проверьте, нет ли ошибок в машине",
            "       * Oстановить бег",
            "* Замените этикетку на дефектнаю этикетку",
            "* очистите принтер и сканер",
            "* Запустите принтер и сканер, если это возможно",
            "* Нажмите продолжить после включения принтера и сканера и проверки их работоспособности"
        ],
        "He": [
            "*אנא בדוק את התווית שהודפסה",
            "*אם התווית נכונה, אנא צלם ושלח לצוות התמיכה",
            "* אם יש שגיאת חוסר התאמה (צפי לתווית אחת אבל נבדקה תווית אחרת):  ",
            "בדוק את גודל הצעד במכונה",
            "בדוק האם ישנם שגיאות במכונה",
            "עצור את הריצה",
            "*החלף את התווית בתווית פגומה",
            "*נקה את המדפסת והסורק",
            "*אתחל את המדפסת והסורק אם יש זמינות",
            "*לחץ על המשך לאחר הפעלת המדפסת והסורק ובדיקת תקינות"
        ]
    },

    "Files Error": {
        "En": [
            "Please make sure all CSV files are closed"
        ],
        "Ru": [
            "Пожалуйста, убедитесь, что все файлы, такие как EXCEL, закрыты"
        ],
        "He": [
            "אנא וודא כי כל הקבצים סגורים כגון אקסל"
        ]
    },
    "Missed Printing": {
        "En": [
             "* Please check printer error and hard reset the printer by the ON/OFF button",
             "* Stop the run and replace all tags from the empty tag, included till the end of the run with faulty tags"
        ],
        "Ru": [
            "*Проверьте ошибку принтера и выполните полную перезагрузку принтера с помощью 	кнопки ВКЛ/ВЫКЛ",
            "*Остановить бег и заменить все этикетки из пустого тега, "
            "включенные до конца прогона, на неисправные этикетки"
        ],
        "He": [
            "* אנא בדוק את השגיאה במדפסת וטפל בה. לאחר מכן אתחל את המדפסת באמצעות מתג ההפעלה הידני.",
            "* עצור את הריצה והחלף את כלל התגים מהתג הריק, כולל, עד סיום הריצה בתגי דמה (ריבוע שחור)"
        ]
    },

    "Printer Error": {
        "En": [
             "Please hard reset the printer by the ON/OFF button and click continue after printer is alive"
        ],
        "Ru": [
            "Пожалуйста, перезагрузить принтер через кнопку включения, "
            "подождите несколько секунд и нажмите продолжить после запуска принтера»"
        ],
        "He": [
            "אנא אתחל את המדפסת על ידי מתג ההפעלה הידני, המתן מספר רגעים ולחץ על המשך לאחר הפעלת המדפסת"
        ]
    },

    "Scanner Error": {
        "En": [
            "Please disconnect and reconnect the scanner and click continue after 5 seconds when it is alive"
        ],
        "Ru": [
            "Пожалуйста, отключите и снова подключите сканер и нажмите продолжить через 5 секунд после его включения"
        ],
        "He": [
            "אנא נתק וחבר מחדש את הסורק ולחץ על המשך 5 שניות לאחר שנדלק"
        ]
    },

    "Arduino Error": {
        "En": [
            "Please disconnect and reconnect Arduino and click continue after 5 second is alive"
        ],
        "Ru": [
            "Пожалуйста, отключите и снова подключите контроллер - "
            "Arduino и нажмите продолжить через 5 секунд после его включения"
        ],
        "He": [
            "אנא נתק וחבר מחדש את הבקר ה-Arduino ולחץ על המשך 5 שניות לאחר שנדלק"
        ]
    },

    "General Error": {
        "En": [
            "* Please make sure there are no empty tags after printer. If there are, please replace with faulty tags",
            "* If running with scanner, please make sure all passed tags were scanned by the scanner. "
            "  If there are, please replace with faulty tags the above external id if relevant",
            "* Please follow error instruction, if error is unclear please contact Wiliot Support"
        ],
        "Ru": [
            "* Проверьте, есть ли после принтера пустые этикетки, если да, замените с дефектные этикетки",
            "* Если машина работает со сканером штрих-кода, убедитесь, что все напечатанные этикетки проверены под ним",
            "  Если есть этикетки, которые не были протестированы - "
            "пожалуйста, замените распечатанные этикетки с дефектные этикетки",
            "* Пожалуйста, следуйте инструкциям по ошибке, если инструкции по ошибке непонятны, обратитесь с Wiliot"
        ],
        "He": [
            "*בדוק האם ישנם תגים ריקים לאחר המדפסת, במידה וכן אנא החלף בתגי דמה",
            "*במידה והמכונה עובדת עם סורק ברקודים, וודא כי כל התגים המודפסים עברו בדיקה תחתיו",
            "במידה וישנם תגים שלא עברו בדיקה - אנא החלף תגים מודפסים בתגי דמה (תגים שלא נבדקו)",
            "*אנא פעל לפי ההוראות השגיאה, אם הוראות השגיאה אינן ברורות, אנא צור קשר עם צוות התמיכה"
        ]
    }
}



"""
  Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

from wiliot_api import ManufacturingClient
from wiliot_core import GetApiKey, InlayTypes
from wiliot_tools.utils.wiliot_gui.wiliot_gui import *
from wiliot_testers.config.unusable_inlays import UnusableInlayTypes
from wiliot_testers.utils.get_version import get_version
from wiliot_testers.tester_utils import open_json, open_json_cache
from wiliot_testers.wiliot_tester_tag_result import *
from wiliot_tools.test_equipment.test_equipment import serial_ports
import serial.tools.list_ports
import logging
import re
import time
import pandas as pd

PRINT_FORMAT_TO_PASS_JOB_NAME = {'SGTIN': 'SGTIN_QR', 'Barcode': 'BARCODE_8', 'prePrint': 'BLANK'}
PRINT_JOB_OPTIONS = tuple(PRINT_FORMAT_TO_PASS_JOB_NAME.values())
PASS_JOB_NUM = 2
TAG_COUNTER_DIGITS = 4
SGTIN_PREFIX = '(01)00850027865010(21)'
CONFIGS_FOLDER = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'configs')


class ConfigDefaults(object):
    """
    contains the default values for the configuration json
    """

    def __init__(self):
        self.printer_defaults = {'TCP_IP': '192.168.6.61', 'TCP_PORT': '3003', 'TCP_BUFFER': '128',
                                 'enableLineSelection': 'Yes', 'enablePrinterAck': 'No',
                                 'printingDuringMovement': 'Yes',
                                 'printingBidirectionalMovement': 'No'}
        self.single_band_gw_defaults = {'energizingPattern': '18', 'timeProfile': '3,7', 'txPower': '3',
                                        'rssiThreshold': '90', 'plDelay': '150'}
        self.dual_band_gw_defaults = {'energizingPattern': '18', 'secondEnergizingPattern': '52', 'timeProfile': '5,10',
                                      'txPower': '3', 'rssiThreshold': '90', 'plDelay': '100'}
        self.external_hw = {"sensorsEnable": "No",
                            "AutoAttenuatorEnable": "No", "attnComport": "AUTO", "attnval": "0",
                            "scannerType": "rtscan", "scannerFlash": True,
                            "isR2R": "yes", "typeR2R": "arduino"}

    def get_printer_defaults(self):
        return self.printer_defaults

    def get_single_band_gw_defaults(self):
        return self.single_band_gw_defaults

    def get_dual_band_gw_defaults(self):
        return self.dual_band_gw_defaults

    def get_external_hw_defaults(self):
        return self.external_hw


class DefaultGUIValues:
    def __init__(self, gui_type):
        if gui_type == 'Main':
            self.default_gui_values = {
                'toPrint': 'No', 'printOffset': '1', 'printingFormat': 'SGTIN',
                'batchName': 'test_reel',
                'testName': 'Ble_Test_Only', 'inlay': InlayTypes.TIKI_099.value,
                'gen': 'Gen2',
                'desiredTags': '9999999', 'desiredPass': '9999999',
                'surface': SurfaceTypes.AIR.value,
                'conversion': ConversionTypes.STANDARD.value,
                'Environment': 'production', 'OwnerId': 'wiliot-ops', 'operator': '', 'conversionLabel': '',
                'QRRead': 'No', 'QRcomport': 'COM3', 'QRoffset': '2',
                'QRtimeout': '200',
                'sensorOffset': '',
                'comments': '', 'maxFailStop': '50', 'maxYieldStop': '40',
                'pass_response_diff': '100', 'pass_response_diff_offset': '9999'}
        elif gui_type == 'Test':
            self.default_gui_values = {'passJobName': 'BARCODE_8', 'passJobNum': 2, 'sgtin': '',
                                       'stringBeforeCounter': '',
                                       'reelNumManually': 'test', 'firstPrintingValue': '0', 'tagLocation': '0',
                                       'tag_reel_location': '0'}
        elif gui_type == 'SGTIN':
            self.default_gui_values = {'passJobName': 'SGTIN_QR', 'passJobNum': 2, 'sgtin': '(01)00850027865010(21)',
                                       'stringBeforeCounter': '',
                                       'reelNumManually': '', 'firstPrintingValue': '0', 'tagLocation': '0',
                                       'tag_reel_location': '0'}
        elif gui_type == 'Barcode':
            self.default_gui_values = {'passJobName': 'BARCODE_8', 'passJobNum': 2, 'sgtin': '',
                                       'stringBeforeCounter': '',
                                       'reelNumManually': '', 'firstPrintingValue': '0', 'tagLocation': '0',
                                       'tag_reel_location': '0'}
        elif gui_type == 'prePrint':
            self.default_gui_values = {'passJobName': 'BLANK', 'passJobNum': 2, 'sgtin': '',
                                       'stringBeforeCounter': '',
                                       'reelNumManually': '', 'firstPrintingValue': '0', 'tagLocation': '0',
                                       'tag_reel_location': '0'}
        else:
            self.default_gui_values = {}


def open_session(test_suite_list):
    """
    gets the user inputs from first GUI
    :return: dictionary of the values
    """

    def check_structure(input_str, gen_in, printing_format, job_name):
        if gen_in.lower() == 'gen2':
            pattern = re.compile(r"^[A-Z0-9]{6}\.[0-9]{2}_[A-F]$")
        elif gen_in.lower() == 'gen3':
            if 'SGTIN' in printing_format.upper() or (job_name is not None and 'SGTIN' in job_name.upper()):
                pattern = re.compile(r"^[a-z0-9]{4}_[A-E]$")
            else:
                pattern = re.compile(r"^[a-z0-9]{3}_[A-E]$")
        else:
            raise Exception('check_structure: gen must be gen2 or gen3')
        if not pattern.match(input_str):
            return False
        return True

    def is_gen3_prod(values_in):
        return values_in['main_gen'].lower() == 'gen3' and 'prod' in values_in['main_Environment'].lower() and \
            values_in['main_toPrint'].lower() == 'yes'

    folder_name = CONFIGS_FOLDER
    file_name = 'gui_inputs_do_not_delete.json'
    gui_inputs_values = open_json(folder_path=folder_name, file_path=os.path.join(folder_name, file_name),
                                  default_values=DefaultGUIValues(gui_type='Main').default_gui_values)
    for key in DefaultGUIValues(gui_type='Main').default_gui_values.keys():
        if key not in gui_inputs_values.keys():
            gui_inputs_values[key] = DefaultGUIValues(gui_type='Main').default_gui_values[key]

    inlay_group = tuple(x.value for x in InlayTypes if x.name not in UnusableInlayTypes.__members__)
    conversion_group = tuple([conv.value for conv in ConversionTypes])
    surface_group = tuple([surf.value for surf in SurfaceTypes])
    im_path = os.path.join(CONFIGS_FOLDER, 'offline_offset.png')
    param_dict = {
        'main': {
            'batchName': {'text': 'Reel Name:', 'value': gui_inputs_values['batchName'], 'tab': 'Main',
                          'group': 'Name'},
            'conversionLabel': {'text': 'conversion Label:', 'value': gui_inputs_values['conversionLabel'],
                                'tab': 'Main', 'group': 'Name'},
            'sameConversionLabel': {'text': 'allow same convertion label for new reel', 'value': False,
                                    'tab': 'Main', 'group': 'Name'},
            'get_reel_id': {'value': 'Get Reel Id', 'widget_type': 'button', 'tab': 'Main', 'group': 'Name'},
            'testName': {'text': 'Test Suite:', 'value': gui_inputs_values['testName'],
                         'widget_type': 'combobox', 'options': tuple(test_suite_list), 'tab': 'Main', 'group': 'test'},
            'toPrint': {'text': 'Printing?', 'value': gui_inputs_values['toPrint'],
                        'widget_type': 'combobox', 'options': ('Yes', 'No'), 'tab': 'Main', 'group': 'test'},
            'QRRead': {'text': 'Scanning?', 'value': gui_inputs_values['QRRead'],
                       'widget_type': 'combobox', 'options': ('Yes', 'No'), 'tab': 'Main', 'group': 'test'},
            'Environment': {'text': 'Environment:', 'value': gui_inputs_values['Environment'], 'group': 'Cloud',
                            'widget_type': 'combobox', 'options': ('test', 'production'), 'tab': 'Main'},
            'OwnerId': {'text': 'Owner Id:', 'value': gui_inputs_values['OwnerId'], 'group': 'Cloud',
                        'widget_type': 'combobox', 'options': ('wiliot-ops', '852213717688'), 'tab': 'Main'},
            'operator': {'text': 'Operator:', 'value': gui_inputs_values['operator'], 'tab': 'Main'},
            'comments': {'text': 'Comments:', 'value': gui_inputs_values['comments'], 'tab': 'Main'},

            'gen': {'text': 'Tag Generation:', 'value': gui_inputs_values['gen'],
                    'widget_type': 'combobox', 'options': ('Gen2', 'Gen3'), 'tab': 'Reel', 'group': 'tag'},
            'inlay': {'text': 'Inlay Type:', 'value': gui_inputs_values['inlay'],
                      'widget_type': 'combobox', 'options': inlay_group, 'tab': 'Reel', 'group': 'tag'},
            'surface': {'text': 'Surface:', 'value': gui_inputs_values['surface'], 'group': 'setup',
                        'widget_type': 'combobox', 'options': surface_group, 'tab': 'Reel'},
            'conversion': {'text': 'Conversion:', 'value': gui_inputs_values['conversion'], 'group': 'setup',
                           'widget_type': 'combobox', 'options': conversion_group, 'tab': 'Reel'},

            'ignoreStop': {'text': 'Ignore Stop Conditions', 'value': False, 'tab': 'Stop Conditions'},
            'maxFailStop': {'text': 'Max failed tags in a row:', 'value': int(gui_inputs_values['maxFailStop']),
                            'tab': 'Stop Conditions'},
            'maxYieldStop': {'text': 'Minimum yield [%]:', 'value': int(gui_inputs_values['maxYieldStop']),
                             'tab': 'Stop Conditions'},
            'desiredTags': {'text': 'Desired amount of tags:', 'value': int(gui_inputs_values['desiredTags']),
                            'tab': 'Stop Conditions'},
            'desiredPass': {'text': 'Desired amount of passed tags:', 'value': int(gui_inputs_values['desiredPass']),
                            'tab': 'Stop Conditions'},

            'printingFormat': {'text': 'Printing Job Format:',
                               'value': gui_inputs_values['printingFormat']
                               if gui_inputs_values['printingFormat'] != 'Test' else 'Barcode',
                               'widget_type': 'combobox', 'options': ('Test', 'SGTIN', 'Barcode', 'prePrint'),
                               'tab': 'Hardware', 'group': 'Printer'},
            'printOffset': {'text': 'Printing Offset:', 'value': int(gui_inputs_values['printOffset']),
                            'widget_type': 'combobox', 'options': tuple(range(1, 6)), 'tab': 'Hardware',
                            'group': 'Printer'},
            'QRcomport': {'text': 'Scanner COM port:',
                          'value': int(gui_inputs_values['QRcomport'])
                          if str(gui_inputs_values['QRcomport']).isdigit() else 0, 'group': 'Scanner',
                          'widget_type': 'combobox', 'options': tuple(range(15)), 'tab': 'Hardware'},
            'QRoffset': {'text': 'Scanner Offset:', 'value': int(gui_inputs_values['QRoffset']),
                         'widget_type': 'combobox', 'options': tuple(range(15)), 'tab': 'Hardware', 'group': 'Scanner'},
            'QRtimeout': {'text': 'Scanner Read Time [ms]:', 'value': int(gui_inputs_values['QRtimeout']),
                          'widget_type': 'combobox', 'options': tuple(range(100, 1000, 100)), 'tab': 'Hardware',
                          'group': 'Scanner'},
            'sensorOffset': {'text': 'Label Detector Sensor Offset:',
                             'value': int(gui_inputs_values['sensorOffset'])
                             if gui_inputs_values['sensorOffset'] else '',
                             'widget_type': 'combobox', 'options': tuple([''] + list(range(15))), 'tab': 'Hardware'},
            "offline_offset_image": {
                "widget_type": "image", "value": im_path, "options": {"width": 550, "height": 300},
                "text": "", 'tab': 'Hardware', 'columnspan': 2},
        },
        'version': {'value': 'Version: {}'.format(get_version()), 'widget_type': 'label'}
    }

    wg = WiliotGui(params_dict=param_dict, do_button_config=False, height_offset=50)

    def update_reel_name_state():
        values_in = wg.get_all_values()
        is_new_name = is_gen3_prod(values_in=values_in)
        for widget_key in ['batchName']:
            wg.update_widget(widget_key=f'main_{widget_key}', disabled=is_new_name)
        for widget_key in ['conversionLabel', 'get_reel_id']:
            wg.update_widget(widget_key=f'main_{widget_key}', disabled=not is_new_name)

    tag_print_config = {}

    def update_conversion_label():
        values_in = wg.get_all_values()
        tag_print_config['conversion_label_changed'] = (values_in["main_conversionLabel"] != gui_inputs_values['conversionLabel']) or values_in["main_sameConversionLabel"]
        if tag_print_config['conversion_label_changed']:
            update_reel_name_str(values_in["main_conversionLabel"])

    def check_conversion_label(label):
        label_parts = label.split('_')
        if len(label_parts) != 5:
            str_out = '\nconversion label must contain 5 fields: inlay, sequence assembly reel, lane id (letter), ' \
                      'conversion date, serial number: (e.g., 169_1_A_22082024_2)'
            return False, str_out
        for i, k in enumerate(['inlay', 'seq_ass_reel', 'lane', 'date', 'sn']):
            tag_print_config[k] = label_parts[i]
        tag_print_config['lane'] = tag_print_config['lane'].upper()
        if not 'A' <= tag_print_config['lane'] <= 'F':
            return False, f'specified lane was: {tag_print_config["lane"]} but labe must be A - F'
        return True, ''

    def update_reel_name_str(user_conversion_label):
        check_conversion_label(user_conversion_label)
        reel_id = tag_print_config.get("stringBeforeCounter", "")
        reel_id = reel_id[-4:] if SGTIN_PREFIX in reel_id else reel_id
        new_str = f'{reel_id}_{tag_print_config.get("lane", "")}'
        wg.update_widget(widget_key='main_batchName', disabled=False)
        wg.update_widget(widget_key='main_batchName', new_value=new_str)
        wg.update_widget(widget_key='main_batchName', disabled=True)

    def update_print_config(print_config):
        print_config['run_printing_gui'] = True
        values_in = wg.get_all_values()
        print_config_out, is_valid, new_run = \
            printing_sgtin_window(env=values_in['main_Environment'], owner_id=values_in['main_OwnerId'],
                                  printing_format=values_in['main_printingFormat'], gen=values_in['main_gen'],
                                  parent=wg.layout)
        tag_print_config['is_new_run'] = new_run
        if is_valid:
            for k, v in print_config_out.items():
                print_config[k] = v
        else:
            print('Generating invalid reel id')
            tag_print_config['stringBeforeCounter'] = 'invalid reel id'
        update_reel_name_str(values_in["main_conversionLabel"])

    def print_offset_change():
        values_in = wg.get_all_values()
        if int(values_in['main_printOffset']) == 0:
            popup_message(msg='Printing offset MUST be larger than 0', bg='red', tk_frame=wg.layout)
            wg.update_widget(widget_key='main_printOffset', new_value=1)
        else:
            popup_message(msg='Printing offset was changed!\nPlease make sure it is the correct offset',
                          bg='yellow', tk_frame=wg.layout)

    def update_stop_conditions_state():
        values_in = wg.get_all_values()
        for widget_key in ['maxFailStop', 'maxYieldStop', 'desiredTags', 'desiredPass']:
            wg.update_widget(widget_key=f'main_{widget_key}', disabled=values_in['main_ignoreStop'])

    def on_submit():
        values_in = wg.get_all_values()
        # check values validity
        err_str = ''
        if not str(values_in['main_desiredTags']).isdigit():
            err_str += f'desiredTags must be a number! {values_in["main_desiredTags"]} is not supported\n'
        if not str(values_in['main_desiredPass']).isdigit():
            err_str += f'desiredPass must be a number! {values_in["main_desiredPass"]} is not supported\n'
        if values_in['main_Environment'] == 'production' and values_in['main_toPrint'].lower() == 'yes':
            if not check_structure(values_in['main_batchName'], values_in['main_gen'],
                                   values_in['main_printingFormat'], tag_print_config.get('passJobName')):
                err_str += 'Reel name error \n Reel name should be: \n'
                if values_in['main_gen'].lower() == 'gen2':
                    err_str += '<6 uppercase alphanumeric chars>.<2 integers minimum 01 maximum 25>_' \
                               '<upper case char minimum A maximum F>'
                elif values_in['main_gen'].lower() == 'gen3':
                    err_str += '<reel id: 3 alphanumeric chars>_<upper case char minimum A maximum E>'
            if values_in['main_gen'].lower() == 'gen3':
                is_label_valid, err_conv = check_conversion_label(values_in["main_conversionLabel"])
                if not is_label_valid:
                    err_str += err_conv
                elif tag_print_config.get('is_new_run', False) and \
                        not tag_print_config.get('conversion_label_changed', False) and not values_in["main_sameConversionLabel"]:
                    err_str += f'\nConversion label was not updated although NEW reel was assembled'
                elif tag_print_config.get('inlay', '') != values_in['main_inlay']:
                    err_str += f'\nUser specified inlay: {values_in["main_inlay"]} is different from the scanned ' \
                               f'conversion label: {values_in["main_conversionLabel"]}'

        if err_str:
            popup_message(msg=err_str, bg='red')
            return

        wg.on_submit()

    update_reel_name_state()
    for widget_event_key in ['gen', 'Environment', 'toPrint']:
        wg.add_event(widget_key=f'main_{widget_event_key}',
                     event_type='<<ComboboxSelected>>', command=lambda args: update_reel_name_state())
    wg.add_event(widget_key='main_conversionLabel',
                 event_type='<FocusOut>', command=lambda args: update_conversion_label())
    wg.add_event(widget_key='main_get_reel_id',
                 event_type='button', command=lambda: update_print_config(print_config=tag_print_config))
    wg.add_event(widget_key='main_printOffset',
                 event_type='<<ComboboxSelected>>', command=lambda args: print_offset_change())
    wg.add_event(widget_key='main_ignoreStop',
                 event_type='button', command=lambda: update_stop_conditions_state())
    wg.button_configs(submit_command=on_submit)
    user_input = wg.run()
    if user_input is None:
        print('User exited the program')
        exit()

    # remove prefix from user input dict
    values = {}
    for k, v in user_input.items():
        values[k.replace('main_', '')] = v

    for key in gui_inputs_values.keys():
        if key not in values.keys():
            values[key] = gui_inputs_values[key]

    # save user input
    with open(os.path.join(folder_name, file_name), 'w') as f:
        json.dump(values, f, indent=2, separators=(", ", ": "), sort_keys=False)

    return values, tag_print_config


def get_printed_value(string_before_the_counter: str, first_counter: str, digits_in_counter=TAG_COUNTER_DIGITS):
    """
    builds the printed value
    :type string_before_the_counter: string
    :param string_before_the_counter: the sg1 Id of the tested reel
    :type digits_in_counter: int
    :param digits_in_counter: amount of digits in the tag counter field (usually 4)
    :type first_counter: string
    :param first_counter: counter of the run first tag
    """
    first_print = str(string_before_the_counter)
    first_print += 'T'
    first_print += str(first_counter).zfill(digits_in_counter)
    is_ok = len(first_counter) <= digits_in_counter
    return first_print, is_ok


def update_tag_counter(current_tag_counter: str, value_to_add: int):
    return str(int(current_tag_counter) + int(value_to_add)).zfill(TAG_COUNTER_DIGITS)


def get_gui_inputs_values(printing_format):
    """
    A function that sends the GUI's input
    """
    folder_name = CONFIGS_FOLDER
    file_name = get_print_user_config_file(printing_format)
    gui_inputs_values = open_json_cache(folder_path=folder_name, file_path=os.path.join(folder_name, file_name),
                                        default_values=DefaultGUIValues(printing_format).default_gui_values)
    return gui_inputs_values


def printing_process_of_test_and_sgtin(window, printing_format, gui_inputs_values, is_new_reel=True):
    """
    A function that does all the work of the printing process
    """
    reel_number = ''
    pass_job_name = None
    is_ok = True
    original_p_format = printing_format
    printing_config = {'pass_job_name': pass_job_name, 'reel_number': reel_number, 'printing_format': printing_format}
    printing_values = {'firstPrintingValue': '', 'reelNumManually': ''}

    def check_first_print(values_in=None, printing_config_in=None):
        if values_in is None:
            values_in = window.get_all_values()
        if printing_config_in is None:
            printing_config_in = printing_config
        printing_config_in['pass_job_name'] = values_in['passJobName']
        try:
            if original_p_format == 'Test':
                printing_config_in["printing_format"] = printing_config_in['pass_job_name']

            if printing_config_in["printing_format"] == 'prePrint':
                if 'T' not in values_in['firstFullExternalId']:
                    window.update_widget(widget_key='-OUTPUT-',
                                         new_value='scanned external id must contains the char T')
                    return
                external_id_list = values_in['firstFullExternalId'].split('T')
                if len(external_id_list) != 2:
                    window.update_widget(widget_key='-OUTPUT-',
                                         new_value='scanned external id must be: <REEL ID>T<TAG COUNT>')
                    return
                if len(external_id_list[1]) != 4:
                    window.update_widget(widget_key='-OUTPUT-',
                                         new_value='Counter should be equal to 4 digits')
                    return
                values_in['firstPrintingValue'] = external_id_list[1]
                values_in['reelNumManually'] = external_id_list[0]
                printing_values['firstPrintingValue'] = values_in['firstPrintingValue']
                printing_values['reelNumManually'] = values_in['reelNumManually']

            if len(str(values_in['firstPrintingValue'])) > TAG_COUNTER_DIGITS:
                window.update_widget(widget_key='-OUTPUT-',
                                     new_value=f"you entered: {values_in['firstPrintingValue']} but length "
                                               f"must be {TAG_COUNTER_DIGITS}")
                return
            if printing_config_in['printing_format'] == 'SGTIN' or 'SGTIN' in values_in['passJobName']:
                printing_config_in['reel_number'] = str(values_in['sgtinNumManually'])
                if not len(str(values_in['sgtinNumManually'])) == 22:
                    window.update_widget(
                        widget_key='-OUTPUT-',
                        new_value=f'SGTIN number is not equal to 22 chars!!\n'
                                  f'The current SGTIN length is: {len(values_in["sgtinNumManually"])}')
                    return
                if not len(str(values_in['reelNumManually'])) == 4:
                    window.update_widget(widget_key='-OUTPUT-',
                                         new_value='Reel number is not equal to 4 chars!!\n'
                                                   'Please enter correct Reel number')
                    return

            if (printing_config_in['printing_format'] == 'Barcode' or 'Barcode' in values_in['passJobName']) and \
                    len(values_in['reelNumManually']) != 3:
                window.update_widget(widget_key='-OUTPUT-',
                                     new_value='Reel number is not equal to 3 chars!!\n'
                                               'Please enter correct Reel number')
                return

            printing_config_in['pass_job_name'] = values_in['passJobName']

            if printing_config_in['printing_format'] == 'SGTIN' or 'SGTIN' in values_in['passJobName']:
                printing_config_in['reel_number'] += str(values_in['reelNumManually'])
            else:
                printing_config_in['reel_number'] = str(values_in['reelNumManually'])
            printing_config_in['first_counter'] = values_in['firstPrintingValue']
            first_print, is_ok = get_printed_value(printing_config_in['reel_number'],
                                                   printing_config_in['first_counter'])
            window.update_widget(widget_key='-OUTPUT-',
                                 new_value='The first tag printing value will be:\n' + first_print)
        except Exception as e:
            window.update_widget(widget_key='-OUTPUT-',
                                 new_value=f'got exception during parsing, please try again: {e}')
            return

        if printing_config_in['printing_format'] == 'prePrint':
            printing_config_in['reel_number'] = str(values_in['reelNumManually'])
            first_print, is_ok = get_printed_value(printing_config_in['reel_number'], values_in['firstPrintingValue'])
            if is_new_reel:
                msg = f'Are you sure the first tag for testing is:\n\n' \
                      f'      {first_print}\n'
            else:
                msg = f'Are you sure the tag above the coupler is:\n\n' \
                      f'      {first_print}\n\n'
                values_in['firstPrintingValue'] = update_tag_counter(values_in['firstPrintingValue'], 1)
                printing_values['firstPrintingValue'] = values_in['firstPrintingValue']
                first_print, _ = get_printed_value(values_in['reelNumManually'], values_in['firstPrintingValue'])
                msg += f'it means that the first tag for testing is:\n\n' \
                       f'      {first_print}\n'
            wg_preprint = WiliotGui(params_dict={'sure': {'value': msg, 'widget_type': 'label'}}, parent=window.layout,
                                    exit_sys_upon_cancel=False)
            out = wg_preprint.run()
            sure_to_submit = out is not None
            if sure_to_submit:
                window.update_widget(widget_key='-OUTPUT-',
                                     new_value='The first tag printing value will be:\n' + first_print)
            else:
                window.update_widget(widget_key='-OUTPUT-',
                                     new_value='')
                return

        return True

    def submit_func(values_in=None, printing_config_in=None):
        if check_first_print(values_in=values_in, printing_config_in=printing_config_in):
            window.on_submit()

    window.button_configs(submit_command=lambda: submit_func(printing_config_in=printing_config))
    window.add_event(widget_key='check_first_print', event_type='button',
                     command=lambda: check_first_print(printing_config_in=printing_config))
    values = window.run()
    if values is None:
        is_ok = False

    v = {}
    pass_job_name = printing_config['pass_job_name']
    reel_number = printing_config['reel_number']

    if is_ok:
        for k, new_v in printing_values.items():
            if new_v:
                values[k] = new_v
        v = {'passJobName': pass_job_name, 'digitsInCounter': TAG_COUNTER_DIGITS, 'passJobNum': PASS_JOB_NUM,
             'firstPrintingValue': values['firstPrintingValue'], 'tagLocation': values['firstPrintingValue'],
             'tag_reel_location': gui_inputs_values['tag_reel_location'],
             'stringBeforeCounter': reel_number}

        data_to_save = {'passJobName': pass_job_name,
                        'passJobNum': PASS_JOB_NUM,
                        'sgtin': values['sgtinNumManually'] if 'sgtinNumManually' in values else '',
                        'reelNumManually': values['reelNumManually'],
                        'firstPrintingValue': values['firstPrintingValue'],
                        'tagLocation': values['firstPrintingValue'],
                        'tag_reel_location': gui_inputs_values['tag_reel_location'],
                        'stringBeforeCounter': reel_number}

        if original_p_format == 'Test':
            v['failJobName'] = pass_job_name
        else:
            v['failJobName'] = 'line_'
            v['printingFormat'] = original_p_format
            data_to_save['printingFormat'] = original_p_format

        folder_name = CONFIGS_FOLDER
        file_name = get_print_user_config_file(original_p_format)
        f = open(os.path.join(folder_name, file_name), "w")
        json.dump(data_to_save, f)
        f.close()
    return v, is_ok


def get_print_user_config_file(printing_format):
    if printing_format.lower() == 'test':
        filename = 'gui_printer_inputs_4_Test_do_not_delete.json'
    elif printing_format in PRINT_FORMAT_TO_PASS_JOB_NAME.keys():
        filename = 'gui_printer_inputs_4_SGTIN_do_not_delete.json'
    else:
        raise Exception(f'unsupported printing format: {printing_format}, valid formats are: sgtin, barcode, test')
    return filename


def printing_test_window(parent=None):
    """
    opens the GUI for user input for test print
    :return: dictionary of user inputs
    """
    printing_format = 'Test'
    gui_inputs_values = get_gui_inputs_values(printing_format)

    if gui_inputs_values['reelNumManually'] == "":
        reel_num = 'test_test_test_test_X_test'
        gui_inputs_values['sgtin'] = reel_num[:22]
        gui_inputs_values['reelNumManually'] = reel_num[22:]
        gui_inputs_values['firstPrintingValue'] = '0'
        gui_inputs_values['tagLocation'] = '0'
        gui_inputs_values['tag_reel_location'] = '0'

    param_dict = {
        'passJobName': {'text': 'Job to print for pass:', 'value': gui_inputs_values['passJobName'],
                        'widget_type': 'combobox', 'options': PRINT_JOB_OPTIONS},
        'firstPrintingValue': {'text': 'What is the first counter number?',
                               'value': gui_inputs_values['firstPrintingValue']},
        'sgtinNumManually': {'text': 'For QR code - what is the SGTIN number?',
                             'value': gui_inputs_values['sgtin']},
        'reelNumManually': {'text': 'For Barcode code - what is the reel number?',
                            'value': gui_inputs_values['reelNumManually']},
        '-OUTPUT-': {'value': '', 'widget_type': 'label', 'options': {'font': FONT_TXT_WARNING}, 'columnspan': 2},
        'check_first_print': {'text': 'Check first print', 'value': '', 'widget_type': 'button'}
    }
    wg = WiliotGui(params_dict=param_dict, parent=parent, do_button_config=False)

    wanted_v, wanted_is_ok = printing_process_of_test_and_sgtin(wg, printing_format, gui_inputs_values)

    return wanted_v, wanted_is_ok


def printing_sgtin_window(env='', owner_id='wiliot-ops', printing_format='SGTIN', gen='Gen2', parent=None):
    """
    opens the GUI for user input for SGTIN print
    :return: dictionary of user inputs
    """
    read_only = False
    gui_inputs_values = get_gui_inputs_values(printing_format)

    # Checking if it's a new reel run and updating the GUI inputs according to this
    params_dict = {
        'space': {'value': '', 'widget_type': 'label'},
        'is_new': {'value': '    New Reel?    \n', 'widget_type': 'label', 'options': {'font': ("Gudea", 20)}}}
    wg_new_run = WiliotGui(params_dict=params_dict,
                           parent=parent, do_button_config=False, exit_sys_upon_cancel=False)
    wg_new_run.button_configs(submit_button_text='Yes', cancel_button_text='No')
    new_run = wg_new_run.run() is not None
    if env == 'production' or env == '':
        env = 'prod'
    if new_run:
        gui_inputs_values['tagLocation'] = '0'
        gui_inputs_values['tag_reel_location'] = '0'
        if printing_format != 'prePrint':
            try:
                logging.info('Receiving data from the cloud, please wait')
                reel_num = get_reel_name_from_cloud_api(env, owner_id, printing_format, gen)
                print(reel_num)
            except Exception:
                logging.warning('Problem with receiving data from cloud')
                raise Exception
        else:
            reel_num = {}
        gui_inputs_values['firstPrintingValue'] = '0000'
        if 'data' in reel_num:
            reel_number = reel_num['data']
            if printing_format == 'SGTIN':
                if len(reel_number) < len(SGTIN_PREFIX):
                    reel_number = SGTIN_PREFIX + reel_number
                gui_inputs_values['sgtin'] = reel_number[:22]
                gui_inputs_values['reelNumManually'] = reel_number[22:26]
            elif printing_format == 'Barcode':
                gui_inputs_values['sgtin'] = ''
                gui_inputs_values['reelNumManually'] = reel_number
            else:
                raise Exception(f'printing_format is not supported: {printing_format}')
            read_only = True
        else:
            gui_inputs_values['sgtin'] = ''
            gui_inputs_values['reelNumManually'] = ''
            read_only = False

    param_dict = {
        'p_format_txt': {'value': str(printing_format), 'widget_type': 'label'},
        'passJobName': {'text': 'Job to print for pass:', 'value': PRINT_FORMAT_TO_PASS_JOB_NAME[printing_format],
                        'widget_type': 'combobox', 'options': PRINT_JOB_OPTIONS},
    }

    if printing_format == 'prePrint':
        scan_str = 'place the first tag before the coupler and scan it' if new_run else \
            'scan the tag above the coupler'
        param_dict['firstFullExternalId'] = {'text': scan_str, 'value': ''}
    else:
        param_dict['firstPrintingValue'] = {'text': 'First counter number',
                                            'value': 0000 if read_only else gui_inputs_values['firstPrintingValue']}
        param_dict['reelNumManually'] = {'text': 'Reel number', 'value': gui_inputs_values['reelNumManually']}

    if printing_format == 'SGTIN':
        param_dict['sgtinNumManually'] = {'text': 'SGTIN number', 'value': gui_inputs_values['sgtin']}

    param_dict['-OUTPUT-'] = {'value': '', 'widget_type': 'label', 'options': {'font': FONT_TXT_WARNING},
                              'columnspan': 2}
    param_dict['check_first_print'] = {'value': 'Check first print', 'widget_type': 'button'}
    param_dict['reel_location'] = {'value': 'Tag Reel Location: {}'.format(gui_inputs_values['tag_reel_location']),
                                   'widget_type': 'label'}

    wg = WiliotGui(params_dict=param_dict, parent=parent, do_button_config=False)
    if read_only:
        wg.update_widget(widget_key='firstPrintingValue',
                         disabled=True)
        if 'sgtinNumManually' in param_dict:
            wg.update_widget(widget_key='sgtinNumManually', disabled=True)

    wanted_v, wanted_is_ok = printing_process_of_test_and_sgtin(wg, printing_format, gui_inputs_values, new_run)

    return wanted_v, wanted_is_ok, new_run


def get_validation_data_for_scanning(parent=None):
    params_dict = {'file_path': {'text': 'You selected scanning without printing.\n'
                                         'Please select a file for scanning validation',
                                 'value': '', 'widget_type': 'file_input'}}
    wg = WiliotGui(params_dict=params_dict, parent=parent)
    user_out = wg.run()
    file_path = user_out['file_path']
    if not file_path:
        raise Exception('no file was selected for scanning without printing')
    if not os.path.isfile(file_path):
        raise Exception(f'The specified file does not exist: {file_path} for scanning without printing')
    df = pd.read_csv(file_path)
    if not {'tag_run_location', 'external_id'}.issubset(df.keys()):
        raise Exception(f'The specified file: {file_path} for scanning without printing '
                        f'MUST contains the following columns: tag_run_location, external_id')
    df = df[['tag_run_location', 'external_id']]
    df.drop_duplicates('tag_run_location', inplace=True)
    df.reset_index(inplace=True)
    if pd.isnull(df['external_id']).all():
        raise Exception(f'The specified file: {file_path} for scanning without printing has empty external_id column')
    if pd.Series(df['tag_run_location'] - df.index != df['tag_run_location'].iloc[0]).any():
        raise Exception(f'The specified file: {file_path} for scanning without printing has missing locations, '
                        f'please select a file with one external id per location')
    return df


def save_screen(tested, passed, yield_, missing_labels, ttfgp_avg, ttfp_max_error=1, responded=None):
    """
    open last GUI
    :type tested: int
    :param tested: amount of tested tags
    :type passed: int
    :param passed: amount of passed tags
    :type yield_: float
    :param yield_: yield in the run
    :type missing_labels: int
    :param missing_labels: amount of missing_labels tags
    :type ttfgp_avg: float or None
    :param ttfgp_avg: average of ttfgp (time to first good packet) in this run
    :return dictionary with the user inputs (should upload, last comments)
    """

    if ttfgp_avg is None or str(ttfgp_avg) == 'nan':
        ttfp_avg_str = ''
        ttfp_avg_color = 'black'
    elif ttfgp_avg < ttfp_max_error:
        ttfp_avg_str = f'Average time to first good packet is OK ({round(ttfgp_avg, 4)} secs)'
        ttfp_avg_color = 'green'
    else:
        ttfp_avg_str = f'Average time to first good packet is too high ({round(ttfgp_avg, 4)} secs)'
        ttfp_avg_color = 'red'

    params_dict = {
        'tested': {'value': f'Tags tested = {tested}', 'widget_type': 'label', 'group': 'Run Statistics'},
        'responded': {'value': f'Tags responded = {responded if responded is not None else 0}', 'widget_type': 'label',
                      'group': 'Run Statistics'},
        'passed': {'value': f'Tags passed = {passed}', 'widget_type': 'label', 'group': 'Run Statistics'},
        'yield': {'value': f'Yield = {round(yield_, 2)}%', 'widget_type': 'label', 'group': 'Run Statistics'},
        'missing': {'value': f'Missing labels = {missing_labels}', 'widget_type': 'label', 'group': 'Run Statistics'},
        'ttfp_avg': {'value': ttfp_avg_str, 'widget_type': 'label', 'group': 'Run Statistics'},
        'upload': {'text': 'Would you like to upload this log to the cloud?', 'value': ('Yes', 'No')},
        'comments': {'text': 'Post run comments:', 'value': ''},
    }

    wg = WiliotGui(params_dict=params_dict, exit_sys_upon_cancel=False)
    wg.update_widget(widget_key='ttfp_avg', color=ttfp_avg_color)
    wg.layout.attributes('-topmost', True)  # pop window up
    wg.layout.attributes('-topmost', False)  # allow user to move wind if needed
    values = wg.run()
    if values is None:
        print("user exited the program, upload did not happen")
        values = {'upload': 'No', 'comments': ''}

    return values


def get_reel_name_from_cloud_api(env, owner_id, printing_format='SGTIN', gen='Gen2'):
    """
    API to receive reel number from cloud (should use it to avoid duplications).
    :return: the reel number (in 0x)
    """
    assert ('testerStationName' in os.environ), 'testerStationName is missing from PC environment variables'
    tester_station_name = os.environ['testerStationName']

    try:
        g = GetApiKey(gui_type='ttk', env=env, owner_id=owner_id)
        api_key = g.get_api_key()
        client = ManufacturingClient(api_key=api_key, logger_=logging.getLogger().name, env=env)
        payload = {"printerId": tester_station_name}
        reel_id_3_char = printing_format == "Barcode"
        n_tries = 0
        while True:
            n_tries += 1
            try:
                reel_id = client.get_reel_id(owner_id, payload, reel_id_3_char, gen)
                return reel_id
            except Exception as e:
                if '409' in e.__str__() and n_tries < 3:
                    print(f'got reel id cloud error 409 - internal conflict, trying again {n_tries}/3...')
                    time.sleep(1)
                else:
                    raise e

    except Exception as e:
        raise Exception(f"An exception occurred at get_reel_name_from_cloud_API: {e}")


"""
    R2R Arduino
"""

ARDUINO_BAUD_RATE = 1000000


class R2rGpio(object):
    """
    class to open and use communication to Arduino on R2R machine
    """

    def __init__(self, logger_name=None):
        """
        initialize params and port
        """
        self.baud_rate = ARDUINO_BAUD_RATE
        self.comport = ''
        self.s = None
        self.connected = False
        self.logger = logging.getLogger(logger_name) if logger_name is not None else None
        ports_list = serial_ports()
        if len(ports_list) == 0:
            raise Exception("no serial ports were found. please check your connections")
        self.connect(ports_list)

    def is_connected(self):
        if self.s is None:
            return False

        if not self.s.isOpen():
            return False

        response = self.query("*IDN?")

        if "Williot R2R GPIO" in response:
            self.connected = True
            if self.logger is None:
                print('R2R: Found ' + response + " Serial Number " + self.query("SER?"))
            else:
                self.logger.info('R2R: Found ' + response + " Serial Number " + self.query("SER?"))
            self.s.flushInput()
            return True
        else:
            self.s.close()
            return False

    def connect(self, ports_list):
        for port in ports_list:
            try:
                self.comport = port
                self.s = serial.Serial(self.comport, self.baud_rate, timeout=0, write_timeout=0)

                if self.is_connected():
                    break

            except (OSError, serial.SerialException):
                pass
            except Exception as e:
                if self.logger is None:
                    print(e)
                else:
                    self.logger.warning(f'R2R: connect: {e}')
        if not self.connected:
            raise Exception('Could NOT connect to the Arduino, please check connections')

    def __del__(self):
        if self.s is not None:
            self.s.close()

    def write(self, cmd):
        """
        Send the input cmd string via COM Socket
        """
        if self.s.isOpen():
            pass
        else:
            self.s.open()

        try:
            self.s.flushInput()
            self.s.write(str.encode(cmd))
        except Exception:
            pass

    def query(self, cmd):
        """
        Send the input cmd string via COM Socket and return the reply string
        :return: massage from arduino (w/o the '\t\n')
        """
        if self.s.isOpen():
            pass
        else:
            self.s.open()
            time.sleep(1)
        self.s.flushInput()
        time.sleep(1)
        try:
            self.s.write(str.encode(cmd))
            time.sleep(2)
            data = self.s.readlines()
            value = data[0].decode("utf-8")
            # Cut the last character as the device returns a null terminated string
            value = value[:-2]
        except Exception:
            value = ''
        return value

    def read(self):
        """
        Send the input cmd string via COM Socket and return the reply string
        :return: massage from arduino (w/o the '\t\n')
        """
        if self.s.isOpen():
            pass
        else:
            self.s.open()
        try:
            timeout = time.time() + 5  # 5 second to receive arudio pulse
            while self.s.in_waiting == 0 and time.time() < timeout:
                pass
            if time.time() > timeout:
                logging.warning('No data received from Arduino')
                self.reconnect()

            data = self.s.readlines()
            self.s.flushInput()
            value = data[0].decode("utf-8")
            # Cut the last character as the device returns a null terminated string
            value = value[:-2]
        except Exception:
            value = ''
        return value

    def gpio_state(self, gpio, state):
        """
        gets the gpio state:
            my_gpio.gpio_state(3, "ON")
               start"on"/stop"off"
            my_gpio.gpio_state(4, "ON")
               enable missing label
        :param gpio: what gpio to write to
        :param state: to what state to transfer (ON / OFF)
        :return: reply from Arduino
        """
        cmd = 'GPIO' + str(gpio) + "_" + state
        replay = self.query(cmd)
        return replay

    def pulse(self, gpio, pulse_duration_ms):
        """
        send a pulse to the r2r machine:
            my_gpio.pulse(1, 1000)
               Pass
            my_gpio.pulse(2, 1000)
               fail
        :param gpio: what gpio to write to
        :param pulse_duration_ms: how long is the pulse
        :return: True if succeeded, False otherwise
        """
        cmd = 'GPIO' + str(gpio) + '_PULSE ' + str(pulse_duration_ms)
        self.write(cmd)
        time.sleep(pulse_duration_ms * 2 / 1000)
        replay = self.read()
        if replay == "Completed Successfully":
            return True
        else:
            return False

    def reconnect(self):
        if self.is_connected():
            logging.info(f'Already connected to Arduino on {self.comport}')
            return True
        try:
            # Attempt to open a connection to the specified comport
            self.s = serial.Serial(self.comport, self.baud_rate, timeout=0, write_timeout=0)
            logging.info(f'Connection to Arduino on {self.comport} reestablished.')
            reconnect_success = True

        except serial.serialutil.SerialException:
            logging.warning(
                f'Connection to {self.comport} failed. Attempting to close the connection and detect Arduino...')
            try:
                self.s.close()
                time.sleep(1)
                self.s = serial.Serial(self.comport, self.baud_rate, timeout=0, write_timeout=0)
                logging.info(f'Connection to Arduino on {self.comport} reestablished.')
                reconnect_success = True

            except Exception as e:
                logging.warning(f"Failed to close connection to {self.comport} {e} - Please restart connection")
                return False

        if reconnect_success:

            if self.is_connected():
                logging.info(
                    'Connection was restablished on COM{} with version {}'.format(self.comport, self.query("SER?")))
                self.s.flushInput()
            else:
                self.s.close()
                logging.warning(f"Failed to close connection to {self.comport} - Please restart connection")
                reconnect_success = False
        return reconnect_success


if __name__ == '__main__':
    # vals = save_screen(tested=100, passed=99, responded=102, yield_=99.54682, ttfgp_avg=1.23, missing_labels=0)
    # open_session(['Hi', 'Bye'])
    # f = get_validation_data_for_scanning()
    # s = printing_test_window()
    # a = printing_sgtin_window(env='test', owner_id='wiliot-ops', printing_format='SGTIN', gen='gen3')
    # rsp = get_reel_name_from_cloud_api(env='test', owner_id='wiliot-ops', printing_format='SGTIN', gen='Gen3')

    v = open_session(['Single Band', 'Dual Band'])
    b = printing_test_window()
    c = printing_sgtin_window(env='test', printing_format='Barcode')
    d = printing_sgtin_window(env='test', printing_format='SGTIN')

    r = R2rGpio()
    open_session(['Single Band', 'Dual Band'])
    pass



import re
import tkinter as tk
from tkinter import filedialog
from datetime import datetime


def two_files_compare_strings(str1, str2):
    lines1 = str1.strip().split('\n')
    lines2 = str2.strip().split('\n')
    min_lines = min(len(lines1), len(lines2))

    output_str = ""
    for i in range(min_lines):
        seconds1 = extract_seconds(lines1[i])
        seconds2 = extract_seconds(lines2[i])
        if seconds1 is not None and seconds2 is not None:
            difference = abs(seconds1 - seconds2)
            if difference > 1:
                output_str += f"{i + 1}: {difference} seconds difference\n"
        else:
            output_str += f"Line {i + 1}: Unable to extract seconds from one or both strings\n"
    return output_str


def two_files_process_file(filepath):
    pattern1 = re.compile(r"(\d{2}/\d{2}/\d{4} \d{2}:\d{2}:\d{2})")
    pattern2 = re.compile(r"(\d{2}/\d{2}/\d{4} \d{2}:\d{2}:\d{2}) (AM|PM)")

    output_str = ""
    try:
        with open(filepath, 'r') as file:
            lines = file.readlines()
            previous_timestamp = None
            for index, line in enumerate(lines, 1):
                if 'msg' not in line and 'Matrix' not in line:
                    continue

                match1 = pattern1.search(line)
                match2 = pattern2.search(line)

                if match1:
                    if 'movement' in line:
                        continue
                    current_timestamp = datetime.strptime(match1.group(1), '%Y-%m-%d %H:%M:%S')
                elif match2:
                    if 'Arduino' in line:
                        continue
                    current_timestamp = datetime.strptime(match2.group(1) + " " + match2.group(2),
                                                          "%m/%d/%Y %I:%M:%S %p")
                else:
                    output_str += f"Could not parse time from line {index}: {line}\n"
                    continue

                if previous_timestamp:
                    delta = (current_timestamp - previous_timestamp).total_seconds()
                    output_str += f"{index - 1}. {int(delta)} seconds\n"
                previous_timestamp = current_timestamp
    except Exception as e:
        output_str += str(e)
    return output_str


def extract_seconds(line):
    match = re.search(r'(\d+(?:\.\d+)?)\s*seconds', line)
    if match:
        return float(match.group(1))
    return None


def compare_strings(str1, str2):
    lines1 = str1.strip().split('\n')
    lines2 = str2.strip().split('\n')
    min_lines = min(len(lines1), len(lines2))

    output_str = ""
    for i in range(min_lines):
        seconds1 = extract_seconds(lines1[i])
        seconds2 = extract_seconds(lines2[i])
        if seconds1 is not None and seconds2 is not None:
            difference = abs(seconds1 - seconds2)
            if difference > 1:
                output_str += f"{i + 1}: {difference} seconds difference\n"
    return output_str


def process_file(filepath):
    output_str = ""
    lines_list = []
    try:
        with open(filepath, 'r') as file:
            lines = file.readlines()
            previous_timestamp = None
            for index, line in enumerate(lines, 1):
                match = re.search(r'(\d{2}/\d{2}/\d{4} \d{2}:\d{2}:\d{2})', line)
                if match:
                    current_timestamp = datetime.strptime(match.group(1), '%m/%d/%Y %H:%M:%S')
                    if previous_timestamp:
                        delta = (current_timestamp - previous_timestamp).total_seconds()
                        output_str += f"{index - 1}. {int(delta)} seconds\n"
                    previous_timestamp = current_timestamp
                    lines_list.append(line)
    except Exception as e:
        output_str += str(e)
    return output_str, lines_list


def post_process_file(lines_list):
    for i in range(len(lines_list) - 1):
        line1 = lines_list[i]
        line2 = lines_list[i + 1]
        match1 = re.search(r'(\d{2}/\d{2}/\d{4} \d{2}:\d{2}:\d{2})', line1)
        match2 = re.search(r'(\d{2}/\d{2}/\d{4} \d{2}:\d{2}:\d{2})', line2)
        if match1 and match2:
            timestamp1 = datetime.strptime(match1.group(1), '%m/%d/%Y %H:%M:%S')
            timestamp2 = datetime.strptime(match2.group(1), '%m/%d/%Y %H:%M:%S')
            delta = (timestamp2 - timestamp1).total_seconds()
            if delta > 50:
                print(
                    f"{i + 1}-{i + 2}: {delta} secs [{timestamp1.strftime('%H:%M:%S')} to "
                    f"{timestamp2.strftime('%H:%M:%S')}]")


def browse_file(file_var):
    filepath = filedialog.askopenfilename()
    file_var.set(filepath)


def main():
    root = tk.Tk()
    root.title('File Comparison')

    yield_run_file = tk.StringVar()
    arduino_run_file = tk.StringVar()

    tk.Label(root, text='Select arduino_run file').pack()
    tk.Entry(root, textvariable=yield_run_file, width=50).pack()
    tk.Button(root, text='Browse', command=lambda: browse_file(yield_run_file)).pack()

    tk.Label(root, text='Select yield_run file').pack()
    tk.Entry(root, textvariable=arduino_run_file, width=50).pack()
    tk.Button(root, text='Browse', command=lambda: browse_file(arduino_run_file)).pack()

    tk.Button(root, text='Compare two Arduino triggers',
              command=lambda: print(compare_strings(two_files_process_file(yield_run_file.get()),
                                                    two_files_process_file(arduino_run_file.get())))).pack()

    tk.Button(root, text='Check machine stops',
              command=lambda: (print(compare_strings(process_file(yield_run_file.get())[0],
                                                     process_file(arduino_run_file.get())[0])),
                               post_process_file(process_file(yield_run_file.get())[1]))).pack()

    root.mainloop()


if __name__ == "__main__":
    main()

